##
# File:  ChemRefDataDbUtils.py
# Date:  24-Jan-2013  J. Westbrook
#
# Updated:
#  4-Feb-2013  jdw generalized for BIRD and chemical component database loading.
# 14-Aug-2013  jdw misc updates
# 11-Nov-2014  jdw add multiprocessing loader for chemical component data -
#  1-Feb-2017  jdw change base class
#
"""
Wrapper for utilities for database loading of chemical reference data content from
repositories and sandboxes.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import sys
import os
import os.path
import time
import scandir
import logging

# try:
#    from itertools import zip_longest
# except ImportError:
#    from itertools import izip_longest as zip_longest

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCc
from wwpdb.utils.db.MyDbSqlGen import MyDbAdminSqlGen
from wwpdb.utils.db.SchemaDefLoader import SchemaDefLoader
from wwpdb.utils.db.MyDbUtil import MyDbQuery

from rcsb.utils.multiproc.MultiProcPoolUtil import MultiProcPoolUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

from wwpdb.utils.db.BirdSchemaDef import BirdSchemaDef
from wwpdb.utils.db.ChemCompSchemaDef import ChemCompSchemaDef

from mmcif_utils.bird.PdbxPrdIo import PdbxPrdIo
from mmcif_utils.bird.PdbxPrdCcIo import PdbxPrdCcIo
from mmcif_utils.bird.PdbxFamilyIo import PdbxFamilyIo
from mmcif_utils.bird.PdbxPrdUtils import PdbxPrdUtils

from mmcif_utils.chemcomp.PdbxChemCompIo import PdbxChemCompIo

# from mmcif.io.IoAdapterPy       import IoAdapterPy
from mmcif.io.IoAdapterCore import IoAdapterCore

from wwpdb.utils.db.MyConnectionBase import MyConnectionBase

from wwpdb.apps.chem_ref_data.utils.OSVersion import OSVersion

logger = logging.getLogger(__name__)


class ChemRefDataDbUtils(MyConnectionBase):
    """Wrapper for utilities for database loading of content from PRD repositories and sandboxes."""

    #

    def __init__(self, reqObj, verbose=False, log=sys.stderr):
        self.__reqObj = reqObj
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        #
        super(ChemRefDataDbUtils, self).__init__(siteId=self.__siteId, verbose=verbose, log=log)
        #
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = verbose
        #
        # Information injected from the request object -
        #
        # self.__topPath = self.__reqObj.getValue("TopPath")
        # self.__topSessionPath = self.__reqObj.getValue("TopSessionPath")
        #
        self.__sObj = self.__reqObj.getSessionObj()
        self.__sessionPath = self.__sObj.getPath()
        # self.__sessionRelativePath = self.__sObj.getRelativePath()
        # self.__sessionId = self.__sObj.getId()
        #

        self.__ioObj = IoAdapterCore(verbose=self.__verbose, log=self.__lfh)
        #
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        self.__cI = ConfigInfo(self.__siteId)
        self.__cIAppCc = ConfigInfoAppCc(self.__siteId)
        self.__sbTopPath = self.__cIAppCc.get_site_refdata_top_cvs_sb_path()
        self.__projName = self.__cI.get("SITE_REFDATA_PROJ_NAME_CC")

    def loadBird(self):
        """Load database containing BIRD PRD and PRD Family data.  Materialized some sequence
        details prior to load.
        """
        startTime = time.time()
        logger.info("+ChemRefDataLoad(loadBird) Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        try:
            birdCachePath = self.__cIAppCc.get_site_prd_cvs_path()
            birdCcCachePath = self.__cIAppCc.get_site_prdcc_cvs_path()
            birdFamilyCachePath = self.__cIAppCc.get_site_family_cvs_path()
            #
            #
            prd = PdbxPrdIo(verbose=self.__verbose, log=self.__lfh)
            prd.setCachePath(birdCachePath)
            pathList = prd.makeDefinitionPathList()
            #
            for pth in pathList:
                prd.setFilePath(pth)
            if self.__verbose:
                logger.info("+ChemRefDataLoad(loadBird) PRD repository read completed")
            #
            prdU = PdbxPrdUtils(prd, verbose=self.__verbose, log=self.__lfh)
            _rD = prdU.getComponentSequences(addCategory=True)  # noqa: F841  might have side effects
            #
            #
            # Load family files
            prdFam = PdbxFamilyIo(verbose=self.__verbose, log=self.__lfh)
            prdFam.setCachePath(birdFamilyCachePath)
            familyPathList = prdFam.makeDefinitionPathList()
            #
            for pth in familyPathList:
                prdFam.setFilePath(pth)
            if self.__verbose:
                logger.info("+ChemRefDataLoad(loadBird) Family repository read completed")
            #
            # Combine containers -
            containerList = prd.getCurrentContainerList()
            containerList.extend(prdFam.getCurrentContainerList())

            # Load prdcc files
            prdCc = PdbxPrdCcIo(verbose=self.__verbose, log=self.__lfh)
            prdCc.setCachePath(birdCcCachePath)
            prdccPathList = prdCc.makeDefinitionPathList()
            #
            for pth in prdccPathList:
                prdCc.setFilePath(pth)
            if self.__verbose:
                logger.info("+ChemRefDataLoad(loadBird) Prdcc repository read completed")
            #
            # Combine containers -
            containerList.extend(prdCc.getCurrentContainerList())

            #
            # Run loader on container list --
            #
            self.setResource(resourceName="PRD")
            ok = self.openConnection()
            if ok:
                sd = BirdSchemaDef()
                self.__schemaCreate(schemaDefObj=sd)

                sdl = SchemaDefLoader(
                    schemaDefObj=sd, ioObj=self.__ioObj, dbCon=self._dbCon, workPath=self.__sessionPath, cleanUp=False, warnings="error", verbose=self.__verbose, log=self.__lfh
                )
                ok = sdl.load(containerList=containerList, loadType="batch-file", deleteOpt="truncate")

                self.closeConnection()
            else:
                if self.__verbose:
                    logger.info("+ChemRefDataLoad(loadBird) database connection failed")

        except:  # noqa: E722 pylint: disable=bare-except
            self.closeConnection()
            logger.exception("loadBird exception...")
            ok = False

        endTime = time.time()
        logger.info("+ChemRefDataLoad(loadBird) Completed at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ok

    def loadChemComp(self):
        startTime = time.time()
        logger.info("+ChemRefDataDbUtils(loadChemComp) Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        try:
            chemCompCachePath = self.__cIAppCc.get_site_cc_cvs_path()
            #
            cc = PdbxChemCompIo(verbose=self.__verbose, log=self.__lfh)
            cc.setCachePath(chemCompCachePath)
            pathList = cc.makeComponentPathList()
            #
            for pth in pathList:
                cc.setFilePath(pth)

            containerList = cc.getCurrentContainerList()
            if self.__verbose:
                logger.info("+ChemRefDataDbUtils(loadChemComp) Reading definitions in chemical component repository read completed length %r", len(containerList))

            # --------------------------------------------------
            # Create schema and run loader on container list --
            #
            self.setResource(resourceName="CC")
            ok = self.openConnection()
            if ok:
                sd = ChemCompSchemaDef()
                self.__schemaCreate(schemaDefObj=sd)
                #
                # sdl=SchemaDefLoader(schemaDefObj=sd,ioObj=self.__ioObj,dbCon=self._dbCon,workPath='.',
                #                cleanUp=False,warnings='error',verbose=self.__verbose,log=self.__lfh)
                sdl = SchemaDefLoader(
                    schemaDefObj=sd, ioObj=self.__ioObj, dbCon=self._dbCon, workPath=self.__sessionPath, cleanUp=False, warnings="error", verbose=self.__verbose, log=self.__lfh
                )
                ok = sdl.load(containerList=containerList, loadType="batch-file", deleteOpt="truncate")
                self.closeConnection()
            else:
                if self.__verbose:
                    logger.info("+ChemRefDataDbUtils(loadChemComp) database connection failed")

        except:  # noqa: E722 pylint: disable=bare-except
            self.closeConnection()
            logger.exception("load ChemComp...")
            ok = False

        endTime = time.time()
        logger.info("+ChemRefDataDbUtils(loadChemComp) Completed at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ok

    def __schemaCreate(self, schemaDefObj):
        """Create and load table schema using schema definition"""
        startTime = time.time()
        logger.info("+ChemRefDataLoad(__schemaCreate) Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        try:
            tableIdList = schemaDefObj.getTableIdList()
            sqlGen = MyDbAdminSqlGen(self.__verbose, self.__lfh)
            sqlL = []
            for tableId in tableIdList:
                tableDefObj = schemaDefObj.getTable(tableId)
                sqlL.extend(sqlGen.createTableSQL(databaseName=schemaDefObj.getDatabaseName(), tableDefObj=tableDefObj))

            if self.__debug:
                logger.debug("+ChemRefDataLoad(__schemaCreate) Schema creation SQL string\n %s", "\n".join(sqlL))

            myQ = MyDbQuery(dbcon=self._dbCon, verbose=self.__verbose, log=self.__lfh)
            #
            # Permit warnings to support "drop table if exists" for missing tables.
            #
            myQ.setWarning("default")
            ret = myQ.sqlCommand(sqlCommandList=sqlL)
            if self.__verbose:
                logger.info("+INFO mysql server returns %r", ret)
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Schema creation...")
            return False

        endTime = time.time()
        logger.info("+ChemRefDataLoad(__schemaCreate) Completed at %s (%.2f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret

    # def __makeSubLists(self, n, iterable):
    #     args = [iter(iterable)] * n
    #     return ([e for e in t if e is not None] for t in zip_longest(*args))

    def makeComponentPathListMulti(self, dataList, procName, optionsD, workingDir):  # pylint: disable=unused-argument
        """Return the list of chemical component definition file paths in the current repository."""
        pathList = []
        for subdir in dataList:
            dd = os.path.join(self.__sbTopPath, self.__projName, subdir)
            for root, _dirs, files in scandir.walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.endswith(".cif") and len(name) <= 9:  # support five character CCID
                        pathList.append(os.path.join(root, name))

        return dataList, pathList, []

    def loadChemCompMulti(self, numProc=8):
        """Create batch load files for all chemical component definition data files - (multiprocessing for file generation only)"""
        if self.__verbose:
            logger.info("Starting")
        startTime = time.time()
        try:
            dataS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
            dataList = [a for a in dataS]

            # Extended CCD support
            ext_ccd = self.__cIAppCc.get_extended_ccd_supp()
            if ext_ccd:
                dataList += [a + b for a in dataS for b in dataS]

            if OSVersion().IsRhel8Like() is False:
                mpu = MultiProcUtil(verbose=True)
                mpu.set(workerObj=self, workerMethod="makeComponentPathListMulti")
                ok, _failList, retLists, _diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            else:
                mppu = MultiProcPoolUtil(verbose=True)
                mppu.set(workerObj=self, workerMethod="makeComponentPathListMulti")
                ok, _failList, retLists, _diagList = mppu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            if self.__verbose:
                logger.info("Path list length %d  in %.2f seconds", len(pathList), endTime0 - startTime)

            ccsd = ChemCompSchemaDef()
            sml = SchemaDefLoader(
                schemaDefObj=ccsd, ioObj=self.__ioObj, dbCon=None, workPath=self.__sessionPath, cleanUp=False, warnings="default", verbose=self.__verbose, log=self.__lfh
            )

            #
            if OSVersion().IsRhel8Like() is False:
                mpu = MultiProcUtil(verbose=True)
                mpu.set(workerObj=sml, workerMethod="makeLoadFilesMulti")
                mpu.setWorkingDir(self.__sessionPath)
                ok, _failList, retLists, _diagList = mpu.runMulti(dataList=pathList, numProc=numProc, numResults=2)
                #
                # containerNameList = retLists[0]
                tList = retLists[1]
            else:
                mppu = MultiProcPoolUtil(verbose=True)
                mppu.set(workerObj=sml, workerMethod="fetchMulti")
                mppu.setWorkingDir(self.__sessionPath)
                ok, _failList, tableDictList, _diagList = mppu.runMulti(dataList=pathList, numProc=numProc, numResults=2)

                # we must first join the retList and merge the dictionaries
                # this must be improved
                tableDict = {}
                for mpresult in tableDictList[1]:
                    for k, v in mpresult.items():
                        tableDict.setdefault(k, [])
                        tableDict[k].extend(v)

                # create the files single-threadily
                tList = sml.export(tableDict=tableDict)
                logger.info("Export list %s", tList)

            if self.__verbose:
                for tId, fn in tList:
                    logger.info("Created table %s load file %s", tId, fn)
            #
            endTime1 = time.time()
            if self.__verbose:
                logger.info("Batch files %r created in %.4f seconds", len(tList), endTime1 - endTime0)

            #
            # --------------------------------------------------
            #               Recreate schema
            #
            self.setResource(resourceName="CC")
            ok = self.openConnection()
            if ok:
                # sd = ChemCompSchemaDef()
                self.__schemaCreate(schemaDefObj=ccsd)
                self.closeConnection()
            else:
                if self.__verbose:
                    logger.info("+ChemRefDataDbUtils(loadChemCompMulti) database connection failed")
                return False
            #
            #
            self.setResource(resourceName="CC")
            ok = self.openConnection()
            #
            sdl = SchemaDefLoader(
                schemaDefObj=ccsd, ioObj=self.__ioObj, dbCon=self._dbCon, workPath=self.__sessionPath, cleanUp=False, warnings="default", verbose=self.__verbose, log=self.__lfh
            )

            ok = sdl.loadBatchFiles(loadList=tList, containerNameList=None, deleteOpt=None)
            self.closeConnection()

            # --------------------------------------------------
            endTime2 = time.time()
            if self.__verbose:
                logger.info("Load completed in %.4f seconds", endTime2 - endTime1)
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Loading...")
            return False

        endTime = time.time()
        logger.info("Completed at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)



##
# File:  BirdReport.py
# Date:  19-Nov-2010
#
# Updates:
#     03-Feb-2012  Jdw  V3 PRD support --
#     31-Jan-2013  Jdw  Integrated within chemical reference data module
#     14-Jun-2017  jdw. refactored add coordinate handling
##
"""
PRD report generator -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import shutil

import logging

from mmcif_utils.bird.PdbxPrdIo import PdbxPrdIo
from mmcif_utils.chemcomp.PdbxChemCompIo import PdbxChemCompIo
from mmcif_utils.bird.PdbxPrdUtils import PdbxPrdUtils
from wwpdb.io.locator.ChemRefPathInfo import ChemRefPathInfo
from wwpdb.apps.chem_ref_data.report.ReportUtils import ReportUtils

logger = logging.getLogger(__name__)


class BirdReport(object):
    """PRD report generator functions."""

    def __init__(self, reqObj, verbose=False, log=sys.stderr):
        """PRD report generator.

        :param `verbose`:  boolean flag to activate verbose logging.
        :param `log`:      stream for logging.

        """
        self.__verbose = verbose
        self.__lfh = log
        #
        self.__reqObj = reqObj
        #
        self.__sObj = self.__reqObj.newSessionObj()
        self.__sessionPath = self.__sObj.getPath()
        self.__sessionRelativePath = self.__sObj.getRelativePath()
        self.__sessionId = self.__sObj.getId()
        #
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        self.__crPI = ChemRefPathInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)

        #
        self.__prdId = None
        self.__prdFilePath = None
        self.__prdCcFilePath = None
        self.__prdFileFormat = "cif"
        #
        #
        self.__rU = ReportUtils(verbose=self.__verbose, log=self.__lfh)
        #

    def setPrdId(self, prdId):
        """Set an existing PRD identifier in archive collection as
        the report target
        """
        self.__prdId = str(prdId.upper())
        return self.__getFilePathFromId(self.__prdId)

    def setFilePath(self, prdFilePath, prdFileFormat, prdId):
        self.__prdFilePath = prdFilePath
        self.__prdFileFormat = prdFileFormat
        if not os.access(self.__prdFilePath, os.R_OK):
            return False
        if prdId is not None:
            self.__prdId = str(prdId).upper()
        #
        return True

    def getFilePath(self):
        return self.__prdFilePath

    def __getFilePathFromId(self, prdId):
        """ """
        self.__prdFilePath = self.__crPI.getFilePath(prdId)
        if not os.access(self.__prdFilePath, os.R_OK):
            return False
        self.__prdFileFormat = "cif"
        return True

    #

    def doReport(self):
        """Return content required to render a PRD report."""
        #
        oD = {}
        oD["dataDict"] = {}
        filePath = self.__prdFilePath
        fileFormat = self.__prdFileFormat
        blockId = self.__prdId
        prdCcFilePath = self.__prdCcFilePath
        #
        logger.info("+BirdReport.doReport()  - Starting doReport()")
        logger.info("+BirdReport.doReport()  - file path    %s", filePath)
        logger.info("+BirdReport.doReport()  - file format  %s", fileFormat)
        logger.info("+BirdReport.doReport()  - block Id     %s", blockId)
        #
        #
        # Make local copies of files -
        #
        (_pth, fileName) = os.path.split(filePath)
        dirPath = os.path.join(self.__sessionPath, "report")
        localPath = os.path.join(dirPath, fileName)
        localRelativePath = os.path.join(self.__sessionRelativePath, "report", fileName)
        if filePath != localPath:
            if not os.access(dirPath, os.F_OK):
                os.makedirs(dirPath)
            shutil.copyfile(filePath, localPath)
            #
            logger.info("+BirdReport.doReport() - Copied input file %s to report session path %s", filePath, localPath)

        #
        # Path context --
        oD["idCode"] = self.__prdId
        oD["filePath"] = filePath
        oD["localPath"] = localPath
        oD["localRelativePath"] = localRelativePath
        oD["sessionId"] = self.__sessionId
        oD["editOpNumber"] = 0
        oD["requestHost"] = self.__reqObj.getValue("request_host")
        oD["imageRelativePath"] = None
        oD["xyzRelativePath"] = None
        #
        pathToImage = None
        try:
            prdR = PdbxPrdIo(verbose=self.__verbose, log=self.__lfh)
            prdR.setFilePath(localPath, prdId=blockId)
            prdR.getPrd()

            prdU = PdbxPrdUtils(prdR, verbose=self.__verbose, log=self.__lfh)
            _rD = prdU.getComponentSequences(addCategory=True)  # noqa: F841  Might have side effects
            #
            if self.__prdId.startswith("PRD_"):
                tD = prdU.getChemCompIds()
                logger.info("For PRD %r CHEMICAL COMPONENT IDS are %r", self.__prdId, tD.items())
                tId = tD[self.__prdId]
                pathToImage = self.__crPI.getFilePath(tId)
                logger.info("ID for IMAGE is %r file path %r", tId, pathToImage)

                imagePath = os.path.join(self.__sessionPath, tId + ".svg")
                ok = self.__rU.makeImagesFromPathList(pathList=[pathToImage], imagePath=imagePath, rows=1, cols=1, imageX=700, imageY=700)
                if ok:
                    imageRelativePath = os.path.join(self.__sessionRelativePath, tId + ".svg")
                    oD["imageRelativePath"] = imageRelativePath
                else:
                    oD["imageRelativePath"] = None
                #
                if self.__verbose:
                    logger.info("+BirdReport.doReport() - PRD %r CC ID %r path to image %r", self.__prdId, tId, pathToImage)

            oD["blockId"] = prdR.getCurrentContainerId()
            if self.__verbose:
                logger.info("+BirdReport.doReport() - category name list %r", prdR.getCurrentCategoryNameList())
            # for catName in prdR.getCategoryList():
            for catName in prdR.getCurrentCategoryNameList():
                oD["dataDict"][catName] = prdR.getCategory(catName=catName)
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Report preparation failed for:  %s", fileName)

        #
        oD["hasExpt"] = False
        oD["hasIdeal"] = False
        if pathToImage is not None:
            try:
                localPrdCcRelativePath = ""
                prdCcFilePath = pathToImage
                if prdCcFilePath is not None:
                    (_pth, prdCcFileName) = os.path.split(prdCcFilePath)
                    dp = os.path.join(self.__sessionPath, "report")
                    localPrdCcPath = os.path.join(dp, prdCcFileName)
                    localPrdCcRelativePath = os.path.join(self.__sessionRelativePath, "report", prdCcFileName)
                    if prdCcFilePath != localPrdCcPath:
                        if not os.access(dp, os.F_OK):
                            os.makedirs(dp)
                    shutil.copyfile(prdCcFilePath, localPrdCcPath)
                    #
                    logger.info("+BirdReport.doReport() - Copied input PRDCC file %s to report session path %s", prdCcFilePath, localPrdCcPath)

                tD = {}
                tD["dataDict"] = {}
                ccR = PdbxChemCompIo(verbose=self.__verbose, log=self.__lfh)
                ccR.setFilePath(localPrdCcPath, compId=None)

                tD["blockId"] = ccR.getCurrentContainerId()
                logger.info("Category name list %r", ccR.getCurrentCategoryNameList())
                for catName in ccR.getCurrentCategoryNameList():
                    tD["dataDict"][catName] = ccR.getCategory(catName=catName)
                hasExpt, hasIdeal = self.__rU.coordinatesExist(tD["dataDict"])
                oD["hasExpt"] = hasExpt
                oD["hasIdeal"] = hasIdeal
                oD["xyzRelativePath"] = localPrdCcRelativePath
            except:  # noqa: E722 pylint: disable=bare-except
                logger.exception("Failing with fileName %r", fileName)

        return oD



##
# File:    MiscUtils.py
# Date:    24-Feb-2013
#
# Updates:
##
"""
Miscellaneous utilites.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

# def editDistance(s1, s2):
#     """
#     Compute the Normalized Damerau-Levenshtein distance between two given
#     strings (s1 and s2)
#     """
#     try:
#         maxLen = max(len(s1), len(s2))
#         return float(maxLen - editDistance(s1,s2))/float(maxLen)
#     except:
#         pass
#     return 0
#
# def editDistance(s1, s2):
#     """
#     Compute the Damerau-Levenshtein distance between two given
#     strings (s1 and s2)
#     """
#     d = {}
#     lenstr1 = len(s1)
#     lenstr2 = len(s2)
#     for i in xrange(-1, lenstr1 + 1):
#         d[(i, -1)] = i + 1
#     for j in xrange(-1, lenstr2 + 1):
#         d[(-1, j)] = j + 1
#
#     for i in xrange(0, lenstr1):
#         for j in xrange(0, lenstr2):
#             if s1[i] == s2[j]:
#                 cost = 0
#             else:
#                 cost = 1
#             d[(i, j)] = min(
#                 d[(i - 1, j)] + 1,  # deletion.
#                 d[(i, j - 1)] + 1,  # insertion
#                 d[(i - 1, j - 1)] + cost,  # substitution
#             )
#             if i > 1 and j > 1 and s1[i] == s2[j - 1] and s1[i - 1] == s2[j]:
#                 d[(i, j)] = min(d[(i, j)], d[i - 2, j - 2] + cost)  # transposition
#
#     return d[lenstr1 - 1, lenstr2 - 1]


def multikeysort(items, columns):
    """
    Sort list of dictionaries on multiple keys -

    Example
    uu = [{"a": 1, "b": 2, "c": "tiger" },
    {"a": 2, "b": 1, "c": "tiger" },
    {"a": 3, "b": 5, "c": "bear" },
    {"a": 4, "b": 4, "c": "tiger" },
    {"a": 5, "b": 1, "c": "bear" }
    ]
    result = multikeysort(uu, ['c', 'b', 'a'])
    """
    from operator import itemgetter

    comparers = [((itemgetter(col[1:].strip()), -1) if col.startswith("-") else (itemgetter(col.strip()), 1)) for col in columns]

    def sign(a, b):
        if a < b:
            return -1
        elif a > b:
            return 1
        else:
            return 0

    def comparer(left, right):
        for fn, mult in comparers:
            result = sign(fn(left), fn(right))
            if result:
                return mult * result
        else:  # pylint: disable=useless-else-on-loop
            return 0

    return sorted(items, cmp=comparer)



#!/usr/bin/env python
# !/opt/wwpdb/bin/python
##
# File:    ChemRefDataDbExec.py
# Author:  jdw
# Date:    13-Jul-2014
# Version: 0.001
#
# Updates:
##
"""
Execuction module for chemical reference data database loading --

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.001"

import sys
import os
import logging
from optparse import OptionParser  # pylint: disable=deprecated-module
from enum import Enum

from wwpdb.apps.chem_ref_data.utils.ChemRefDataDbUtils import ChemRefDataDbUtils
from wwpdb.apps.chem_ref_data.utils.ChemRefDataCvsUtils import ChemRefDataCvsUtils
from wwpdb.apps.chem_ref_data.utils.ChemRefDataMiscUtils import ChemRefDataMiscUtils
from wwpdb.utils.config.ConfigInfo import getSiteId
from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCommon
from wwpdb.utils.session.WebRequest import InputRequest

logger = logging.getLogger()


class ChemRefDb(Enum):
    CC = "CC"
    PRD = "PRD"

    def __str__(self):
        return self.value


class ChemRefDataDbExec(object):
    def __init__(self, defSiteId="WWWDPB_INTERNAL_RU", sessionId=None, verbose=True, log=sys.stderr):
        self.__lfh = log
        self.__verbose = verbose
        self.__setup(defSiteId=defSiteId, sessionId=sessionId)

    def __setup(self, defSiteId=None, sessionId=None):
        """Simulate the web application environment for managing session storage of  temporaty data files."""
        self.__siteId = getSiteId(defaultSiteId=defSiteId)
        # print(self.__siteId)
        #
        self.__cICommon = ConfigInfoAppCommon(self.__siteId)
        self.__topPath = self.__cICommon.get_site_web_apps_top_path()
        self.__topSessionPath = self.__cICommon.get_site_web_apps_top_sessions_path()
        #
        self.__reqObj = InputRequest({}, verbose=self.__verbose, log=self.__lfh)
        self.__reqObj.setValue("TopSessionPath", self.__topSessionPath)
        self.__reqObj.setValue("TopPath", self.__topPath)
        self.__reqObj.setValue("WWPDB_SITE_ID", self.__siteId)
        #
        # self.__reqObj.setValue("SITE_REFDATA_DB_USER", os.environ["SITE_REFDATA_DB_USER"])
        # self.__reqObj.setValue("SITE_REFDATA_DB_PASSWORD", os.environ["SITE_REFDATA_DB_PASSWORD"])

        # self.__reqObj.setValue("SITE_REFDATA_CVS_USER", os.environ["SITE_REFDATA_CVS_USER"])
        # self.__reqObj.setValue("SITE_REFDATA_CVS_PASSWORD", os.environ["SITE_REFDATA_CVS_PASSWORD"])

        os.environ["WWPDB_SITE_ID"] = self.__siteId
        if sessionId is not None:
            self.__reqObj.setValue("sessionid", sessionId)

        # We need to create a session obj (with subdir). Do not remove
        _sessionObj = self.__reqObj.newSessionObj()  # noqa: F841
        # self.__reqObj.printIt(ofh=self.__lfh)
        #

    def doLoadChemComp(self, rptPath=None):  # pylint: disable=unused-argument
        """ """
        try:
            dbu = ChemRefDataDbUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            dbu.loadChemComp()
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed in doLoadChemComp")

    def doLoadChemCompMulti(self, numProc):
        """ """
        try:
            dbu = ChemRefDataDbUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok = dbu.loadChemCompMulti(numProc)
            if ok:
                return True
            else:
                logger.info("load CCD failed")
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed load CCD")
        return False

    def doLoadBird(self, rptPath=None):  # pylint: disable=unused-argument
        """ """
        try:
            dbu = ChemRefDataDbUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok = dbu.loadBird()
            if ok:
                return True
            else:
                logger.info("load PRD failed")
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed doLoadBird")
        return False

    def doCheckoutChemComp(self):
        """ """
        try:
            cvsu = ChemRefDataCvsUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            sandbox_path = cvsu.getSandBoxTopPath()
            if sandbox_path:
                if not os.path.exists(sandbox_path):
                    os.makedirs(sandbox_path)
            ok, _textList = cvsu.checkoutChemCompSerial()
            if ok:
                return True
            else:
                logger.info("checkout CCD CVS failed")
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed doCheckoutChemComp")
        return False

    def doCheckoutPRD(self):
        try:
            cvsu = ChemRefDataCvsUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            sandbox_path = cvsu.getSandBoxTopPath()
            if sandbox_path:
                if not os.path.exists(sandbox_path):
                    os.makedirs(sandbox_path)
            ok, _textList = cvsu.checkoutPRDSerial()
            if ok:
                return True
            else:
                logger.info("checkout PRD CVS failed")
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("doCheckoutPRD")
        return False

    def doSyncChemComp(self, numProc):
        """ """
        try:
            cvsu = ChemRefDataCvsUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            sandbox_path = cvsu.getSandBoxTopPath()
            if sandbox_path:
                if not os.path.exists(sandbox_path):
                    logger.info("sandbox path %s does not exist - running checkout", sandbox_path)
                    return self.doCheckoutChemComp()
                ok, _diag_list = cvsu.syncChemComp(numProc)
                if ok:
                    return True
                else:
                    logger.info("CVS update failed for CCD")
                    return False
            else:
                logger.info("sandbox path is None - exiting")
                return False
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("doSyncChemComp")
        return False

    def doSyncBird(self):
        """ """
        try:
            cvsu = ChemRefDataCvsUtils(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            sandbox_path = cvsu.getSandBoxTopPath()
            if sandbox_path:
                if not os.path.exists(sandbox_path):
                    logger.info("sandbox path %s does not exist - running checkout", sandbox_path)
                    return self.doCheckoutPRD()
                ok, _textList = cvsu.syncBird()
                if ok:
                    return True
                else:
                    logger.info("CVS update failed for PRD")
                    return False
            else:
                logger.info("sandbox path is None - exiting")
                return False
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Sandbox update")

    def doUpdateSupportFiles(self):
        """Update chemical component definition file idList, pathList, concatenated dictionary,
        serialized dictionary, dictionary search index, and several Python serialized support files.
        """
        logger.info("Starting")
        try:
            mu = ChemRefDataMiscUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok1 = mu.updateChemCompSupportFiles()
            if not ok1:
                logger.info("updateChemCompSupportFiles failed")
            ok2 = mu.updateChemCompPySupportFiles()
            if not ok2:
                logger.info("updateChemCompPySupportFiles failed")
            ok3 = mu.updatePrdSupportFiles()
            if not ok3:
                logger.info("updatePrdSupportFiles failed")
            return ok1 and ok2 and ok3
        except:  # noqa: E722 pylint: disable=bare-except
            logger.info("Failed")
            logger.exception("Failed in doUpdateSupportFiles")
            return False

    def run_setup_process(self, numProc=8):
        ok1 = self.doCheckoutChemComp()
        ok2 = self.doCheckoutPRD()
        ok3 = self.doLoadChemCompMulti(numProc=numProc)
        ok4 = self.doLoadBird()
        ok5 = self.doUpdateSupportFiles()
        return ok1 and ok2 and ok3 and ok4 and ok5

    def run_update_process(self, numProc=8):
        ok1 = self.doSyncChemComp(numProc=numProc)
        ok2 = self.doSyncBird()
        ok3 = self.doLoadChemCompMulti(numProc=numProc)
        ok4 = self.doLoadBird()
        ok5 = self.doUpdateSupportFiles()

        # hard code syncing CCD CVS to True as it often fails, but really works
        ok1 = True
        return ok1 and ok2 and ok3 and ok4 and ok5


def main():
    FORMAT = "[%(levelname)s]-%(module)s.%(funcName)s: %(message)s"
    logging.basicConfig(format=FORMAT)
    logger.setLevel(logging.INFO)

    usage = "usage: %prog [options]"
    parser = OptionParser(usage)

    parser.add_option("--load", dest="load", action="store_true", default=False, help="Load database from repository sandbox")
    parser.add_option("--sync", dest="sync", action="store_true", default=False, help="Synchronize repository sandbox")
    parser.add_option("--checkout", dest="checkout", action="store_true", default=False, help="Checkout repository into sandbox")
    parser.add_option("--update", dest="update", action="store_true", default=False, help="Update support files from repository sandbox")

    parser.add_option("--db", dest="db", default="PRD", help="Database to load (CC,PRD)")
    parser.add_option("--run_setup", action="store_true", help="Run setup for CCD and PRD")
    parser.add_option("--run_update", action="store_true", help="Run update for CCD and PRD")

    parser.add_option("--numproc", type=int, dest="numProc", default=8, help="Number of processors to engage.")

    parser.add_option("-v", "--verbose", default=False, action="store_true", dest="verbose")
    (options, _args) = parser.parse_args()

    ok = True

    crx = ChemRefDataDbExec(defSiteId="WWWDPB_INTERNAL_RU", sessionId=None, verbose=options.verbose, log=sys.stderr)

    if options.checkout:
        if options.db == "CC":
            ok = crx.doCheckoutChemComp()
        if options.db == "PRD":
            ok = crx.doCheckoutPRD()

    if options.sync:
        if options.db == "CC":
            ok = crx.doSyncChemComp(options.numProc)
            # this often fails even though it works
            ok = True
        elif options.db == "PRD":
            ok = crx.doSyncBird()

    if options.load:
        if options.db == "CC":
            ok = crx.doLoadChemCompMulti(options.numProc)
        elif options.db == "PRD":
            ok = crx.doLoadBird()

    if options.update:
        ok = crx.doUpdateSupportFiles()

    if options.run_setup:
        ok = crx.run_setup_process(numProc=options.numProc)

    if options.run_update:
        ok = crx.run_update_process(numProc=options.numProc)

    if not ok:
        logging.error("task failed")
        sys.exit(1)


if __name__ == "__main__":
    main()



##
# File:  ChemRefSearchDepictBootstrap.py
# Date:  18-Feb-2013 jdw
#
# Updates:
#    28-Apr-2017  Replace server generated tables wioth BootstrapTable rendering -
#
##
"""
Create HTML depictions for simple chemical reference searches using Bootstrap CSS framework -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import random

#
import logging

# Only for full page rendering -
from wwpdb.apps.chem_ref_data.depict.ChemRefDataDepictBootstrap import ChemRefDataDepictBootstrap

logger = logging.getLogger(__name__)


class ChemRefSearchDepictBootstrap(ChemRefDataDepictBootstrap):
    """Create HTML depictions for simple chemical reference searches using Bootstrap CSS framework -"""

    def __init__(self, includePath=None, verbose=False, log=sys.stderr):
        """

        :param `verbose`:  boolean flag to activate verbose logging.
        :param `log`:      stream for logging.

        """
        super(ChemRefSearchDepictBootstrap, self).__init__(includePath, verbose, log)
        #
        #

    def doRenderPage(self, d=None):
        """Render complete HTML page -"""
        #
        if d is None:
            d = {}
        oL = []
        oL.extend(self.appPageTop(title="Chemical Reference Data Management Module", cssPathList=[]))
        #
        oL.extend(self.doRenderResults(d))
        #
        oL.extend(self.appPageBottom(scriptPathList=[]))
        return oL

    def doAltRenderCollapsable(self, d=None, searchName="Search results", compareType=None, searchTarget=None):
        """Render in collapsable section:

        <
        """
        if d is None:
            d = {}
        oL = []
        if d["count"] < 1:
            return oL
        #
        if "compareType" in d and "searchType" in d:
            myTitle = "Results for: %s <i>%s</i> %s &nbsp;&nbsp; " % (searchName, d["compareType"], d["searchTarget"])
        elif compareType and searchTarget:
            myTitle = "Results for: %s <i>%s</i> %s &nbsp;&nbsp; " % (searchName, compareType, searchTarget)
        else:
            myTitle = "%s for target = %s &nbsp;&nbsp; " % (searchName, d["searchTarget"])

        panelTitle = myTitle
        try:
            oL.extend(self.doAltRenderResults(d, title=panelTitle))
        except Exception as e:
            logger.exception("Failing %s", str(e))

        return oL

    def __panelBegin(self, panelTitle, state="collapsed", toggleStyle="href"):
        """Return the prologue for a collapsed panel - BS3+/FA4"""
        # state controls whether the panel is initially expanded or collapsed
        #
        if state == "collapsed":
            panelState = ("", "collapsed")
        else:
            panelState = ("in", "")

        #
        oL = []
        idPrefix = "subpanel-" + str(random.randint(0, 100000))
        idPanel = idPrefix + "-psec-0"
        idPanelHeader = idPrefix + "-phead-0"
        idPanelCollapsable = idPrefix + "-pcol-0"
        #
        #  Not using this level --
        # oL.append('<div class="panel-group" id="%s" role="tablist" aria-multiselectable="true">' % idPanelGroup)
        # -------------------------
        oL.append('  <div id="%s" class="panel panel-default results-section">' % idPanel)

        #
        oL.append('     <div class="panel-heading" role="tab" id="%s">' % idPanelHeader)
        # add close/dismiss --
        oL.append(
            '         <button type="button" class="close" data-target="#%s" data-dismiss="alert"> <span aria-hidden="true"> <i class="fa fa-times"></i></span><span class="sr-only">Close</span></button>'  # noqa: E501
            % idPanel
        )

        oL.append('        <div class="panel-title">')

        if toggleStyle == "href":
            oL.append(
                '          <a role="button" data-toggle="collapse"  href="#%s" aria-expanded="true" aria-controls="%s" class="%s"><i class="chevron fa fa-fw" ></i></a>%s'
                % (idPanelCollapsable, idPanelCollapsable, panelState[1], panelTitle)
            )
        elif toggleStyle == "chevron":
            oL.append(
                '          <a role="button" data-toggle="collapse" href="#%s" aria-expanded="true" aria-controls="%s" class="%s">%s <i class="chevron fa fa-fw" ></i></a>'
                % (idPanelCollapsable, idPanelCollapsable, panelState[1], panelTitle)
            )
        elif toggleStyle == "button":
            oL.append("<span>%s</span>" % panelTitle)
            buttonTitle = "Toggle view"
            oL.append(
                '<button class="btn btn-primary btn-xs" type="button" data-toggle="collapse" data-target="#%s" aria-expanded="false" aria-controls="%s">%s</button>'
                % (idPanelCollapsable, idPanelCollapsable, buttonTitle)
            )

        oL.append("        </div>")

        oL.append("     </div>")

        oL.append('     <div id="%s" class="panel-collapse collapse %s" role="tabpanel" aria-labelledby="%s">' % (idPanelCollapsable, panelState[0], idPanelHeader))
        oL.append('        <div class="panel-body">')
        return oL

    def __panelEnd(self):
        """ """
        oL = []
        # oL.append('        </div>')
        oL.append("     </div>")
        oL.append("  </div> <!-- end of panel -->")
        # -------------------------
        return oL

    def __filterResultSet(self, rD):
        """Add markup to selected column types -"""
        #
        for result in rD["resultlist"]:
            for cName in rD["columnList"]:
                if cName in result and result[cName]:
                    if cName in ["Family ID", "PRD ID"]:
                        result[cName] = '<a class="app-ref-report"  href="#">%s</a>' % result[cName]
                    elif cName in ["CC ID"]:
                        rpt = '<a class="app-ref-report"  href="#">%s</a>' % result[cName]
                        le = '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>' % result[cName]
                        result[cName] = "%s &nbsp; %s" % (rpt, le)
                    elif cName in ["ID"]:
                        if result[cName].startswith("PRD_") or result[cName].startswith("FAM_"):
                            result[cName] = '<a class="app-ref-report"  href="#">%s</a>' % result[cName]
                        else:
                            rpt = '<a class="app-ref-report"  href="#">%s</a>' % result[cName]
                            le = (
                                '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>' % result[cName]
                            )
                            result[cName] = "%s &nbsp; %s" % (rpt, le)
        return rD

    def __bootstrapTableTemplate(
        self,
        colNameTupList,
        columnWidthList,
        title="Search Results",  # pylint: disable=unused-argument
        tableId="chemref-bst-result-table-1",
        divId="chemref-bst-result-div-1",
        pFunc="pagerFunc",
    ):
        """
        Create table template for BootstrapTable parameters --

        colNameTupList = [(colkey,colDisplayName), ... ]
        pagerFunc = <name of JS pager method >

        """
        tableH = """<div id="%s" class="row">
           <table id="%s" data-classes="table  table-bordered table-condensed table-striped"
           data-sort-name="my_data_count"
           data-sort-order="desc"
           data-search="true"
           data-show-refresh="true"
           data-show-toggle="true"
           data-show-columns="true"
           data-query-params="%s"
           data-pagination="true"
           data-page-size="20"
           data-page-list="[20,50,150]"
           data-show-export="true"
           data-export-types="['csv','excel']"
           data-show-refresh="true"
           data-striped="true"
                     data-show-toggle="true">""" % (
            divId,
            tableId,
            pFunc,
        )
        # omit
        #            data-height="600"
        #
        oL = []
        oL.append(tableH)

        # oL.append('<caption class="text-left"><strong> &nbsp;&nbsp; %s</strong></caption>' % title)
        oL.append("<thead><tr>")
        for ii, tup in enumerate(colNameTupList):
            cw = columnWidthList[ii]
            oL.append('<th data-field="%s" data-width="%s" data-sortable="true">%s</th>' % (tup[0], cw, tup[1]))
        oL.append("</tr></thead></table>")
        oL.append("</div><!-- end or result container -->")

        return oL

    def doBsTableRenderCollapsable(self, d=None, searchName="Search results", compareType=None, searchTarget=None):
        """Render in collapsable section using Bootstrap table style --"""
        oL = []
        if d is None:
            d = {}

        if d["count"] < 1:
            return oL
        #
        if "compareType" in d and "searchType" in d:
            myTitle = "Results for: %s <i>%s</i> %s &nbsp;&nbsp; " % (searchName, d["compareType"], d["searchTarget"])
        elif compareType and searchTarget:
            myTitle = "Results for: %s <i>%s</i> %s &nbsp;&nbsp; " % (searchName, compareType, searchTarget)
        else:
            myTitle = "%s for target = %s &nbsp;&nbsp; " % (searchName, d["searchTarget"])

        panelTitle = myTitle
        logger.info("myTitle is %r", myTitle)
        tableDataD = {}
        try:
            tableDataD = self.doBsTableRenderResults(d, title=panelTitle)
        except Exception as e:
            logger.exception("Failing %s", str(e))

        return tableDataD

    def doBsTableRenderResults(self, d=None, title="Search Results Summary"):
        """ """
        if d is None:
            d = {}
        tableDataD = {}
        for resultSetId in range(1, len(d["resultDictionary"]) + 1):
            rD = d["resultDictionary"][resultSetId]
            if len(rD["resultlist"]) < 1:
                continue
            #
            if ("displayTitle" in rD) and (len(rD["displayTitle"]) > 0):
                t = rD["displayTitle"]
                # t = title + ' (' + rD['displayTitle'] + ')'
            else:
                t = title
            #
            #
            oL = []
            tD = {}
            oL.extend(self.__panelBegin(t, state="collapsed", toggleStyle="href"))
            #
            colNameTupList = [(c, c) for c in rD["columnList"]]
            #
            idSuffix = "acc" + str(random.randint(0, 100000))
            tD["resultSetContainerId"] = "chemref-bst-result-div-%s" % idSuffix
            tD["resultSetTableId"] = "chemref-bst-result-table-%s" % idSuffix
            #
            oL.extend(
                self.__bootstrapTableTemplate(colNameTupList, rD["columnWidthList"], title=t, tableId=tD["resultSetTableId"], divId=tD["resultSetContainerId"], pFunc="pagerFunc")
            )
            #
            rD = self.__filterResultSet(rD)
            tD["resultSetTableData"] = rD["resultlist"]
            tD["resultSetId"] = resultSetId
            #
            # rD['resultlist']  [{rowdict},{rowdict},...]
            # rD['columnList']  display column = col key
            #
            oL.extend(self.__panelEnd())
            #
            tD["resultSetTableTemplate"] = "\n".join(oL)
            tableDataD[resultSetId] = tD

        return tableDataD

    def doAltRenderResults(self, d=None, title="Search Results Summary"):
        """Render in HTML"""
        #
        if d is None:
            d = {}

        # sp = '&nbsp;&nbsp;'
        oL = []
        #
        #
        # oL.append('<h4>%s</h4>' % title)

        # if False:
        #     oL.append('<h4>%s</h4>' % title)
        #     oL.append('<p>')
        #     oL.append('Search type:   <strong>%s</strong> ' % d['searchType'])
        #     oL.append('%s' % sp)
        #     oL.append('Search target: <strong>%s</strong> ' % d['searchTarget'])
        #     oL.append('%s' % sp)
        #     oL.append('Comparison type: <strong>%s</strong> ' % d['searchOp'])
        #     oL.append('%s' % sp)
        #     oL.append('Result count: <strong>%r</strong> ' % d['count'])
        #     oL.append('</p>')
        #     oL.append('<br />')

        #
        for resultSetId in range(1, len(d["resultDictionary"]) + 1):
            rD = d["resultDictionary"][resultSetId]
            if len(rD["resultlist"]) < 1:
                continue
            # -- Wrapper
            if ("displayTitle" in rD) and (len(rD["displayTitle"]) > 0):
                t = title + " (" + rD["displayTitle"] + ")"
            else:
                t = title
            oL.extend(self.__panelBegin(t, state="collapsed", toggleStyle="href"))
            # -- Wrapper
            # oL.append('<h4>%s</h4>' % rD['displayTitle'])
            oL.append('<table class="table  table-bordered table-condensed table-striped">')

            for cWidth in rD["columnWidthList"]:
                oL.append(' <col width="%s" />' % cWidth)

            oL.append("<tr>")
            for cName in rD["columnList"]:
                oL.append(" <th>%s</th>" % cName)
            oL.append("</tr>")
            #
            # Row data -
            #
            i = 0
            for result in rD["resultlist"]:
                i += 1
                oL.append("<tr>")
                for cName in rD["columnList"]:
                    if cName in result and result[cName]:
                        if cName == "Family ID":
                            oL.append('          <td><a class="app-ref-report"  href="#">%s</a></td>' % (result[cName]))
                        elif cName == "PRD ID":
                            oL.append('          <td><a class="app-ref-report"   href="#">%s</a></td>' % (result[cName]))
                        elif (cName == "CC ID") or (cName == "ID"):
                            rpt = '<a class="app-ref-report"  href="#">%s</a>' % result[cName]
                            le = (
                                '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>' % result[cName]
                            )
                            oL.append("          <td>%s &nbsp; %s</td>" % (rpt, le))

                        else:
                            oL.append("          <td>%s</td>" % result[cName])
                    else:
                        oL.append("          <td>&nbsp;</td>")
                oL.append("       </td>")
                oL.append("     </tr>")
            oL.append("</table>")
            # -- Wrapper
            oL.extend(self.__panelEnd())
            # --- Wrapper
            #

        # -------------------------------------

        return oL

    def doRenderCollapsable(self, d=None, title="Search results", compareType=None, searchTarget=None):
        """Render in collapsable section -"""
        oL = []

        if d is None:
            d = {}

        if d["count"] < 1:
            return oL
        idPrefix = "acc" + str(random.randint(0, 100000))
        idTop = idPrefix + "-top"

        #
        active = ""
        if compareType and searchTarget:
            myTitle = "Results for: %s <i>%s</i> %s &nbsp;&nbsp; " % (title, compareType, searchTarget)
        else:
            myTitle = "%s for target = %s &nbsp;&nbsp; " % (title, d["searchTarget"])

        idSection = idPrefix + "-sec-0"
        oL.append('<div class="accordion" id="%s">' % idTop)

        oL.append('<div class="accordion-group">')
        oL.append('<div class="accordion-heading">')
        oL.append('<span>%s</span><a class="accordion-toggle" data-toggle="collapse" data-parent="#%s" href="#%s">(toggle view results)</a>' % (myTitle, idTop, idSection))
        oL.append("</div>")
        oL.append('<div id="%s" class="accordion-body collapse %s">' % (idSection, active))
        oL.append('<div  class="accordion-inner">')

        oL.extend(self.doRenderResults(d))

        oL.append("</div>")
        oL.append("</div> <!-- end of accordion group -->")
        oL.append("</div> <!-- end of accordion -->")

        return oL

    def doRenderResults(self, d=None, title="Search Results Summary"):
        """Render in HTML"""
        #
        if d is None:
            d = {}

        sp = "&nbsp;&nbsp;"
        requestHost = "localhost"

        oL = []
        #
        #
        oL.append("<h4>%s</h4>" % title)
        oL.append("<p>")
        oL.append("Search type:   <strong>%s</strong> " % d["searchType"])
        oL.append("%s" % sp)
        oL.append("Search target: <strong>%s</strong> " % d["searchTarget"])
        oL.append("%s" % sp)
        oL.append("Comparison type: <strong>%s</strong> " % d["searchOp"])
        oL.append("%s" % sp)
        oL.append("Result count: <strong>%r</strong> " % d["count"])
        oL.append("</p>")
        oL.append("<br />")

        #
        for resultSetId in range(1, len(d["resultDictionary"]) + 1):
            rD = d["resultDictionary"][resultSetId]

            if len(rD["resultlist"]) < 1:
                continue

            oL.append("<br /><h4>%s</h4>" % rD["displayTitle"])

            oL.append('<table class="table  table-bordered table-condensed table-striped">')

            for cWidth in rD["columnWidthList"]:
                oL.append('	<col width="%s" />' % cWidth)

            oL.append("<tr>")
            for cName in rD["columnList"]:
                oL.append("	<th>%s</th>" % cName)
            oL.append("</tr>")
            #
            # Row data -
            #
            i = 0
            for result in rD["resultlist"]:
                i += 1
                oL.append("<tr>")
                for cName in rD["columnList"]:
                    if cName in result and result[cName]:
                        if cName == "Family ID":
                            oL.append(
                                '          <td><a class="app-ref-report" target="_blank" href="http://%s/service/chemref/bird/report?prdId=%s&format=%s">%s</a></td>'
                                % (requestHost, result[cName], "html", result[cName])
                            )
                        elif cName == "PRD ID":
                            oL.append(
                                '          <td><a class="app-ref-report"  target="_blank" href="http://%s/service/chemref/bird/report?prdId=%s&format=%s">%s</a></td>'
                                % (requestHost, result[cName], "html", result[cName])
                            )
                        elif (cName == "CC ID") or (cName == "ID"):
                            rpt = '<a class="app-ref-report" target="_blank" href="#">%s</a>' % result[cName]
                            le = (
                                '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>' % result[cName]
                            )
                            oL.append("          <td>%s &nbsp; %s</td>" % (rpt, le))

                        else:
                            oL.append("          <td>%s</td>" % result[cName])
                    else:
                        oL.append("          <td>&nbsp;</td>")
                oL.append("  	  </td>")
                oL.append("     </tr>")
            oL.append("</table>")
            #

        # -------------------------------------

        return oL



##
# File:  ChemCompReport.py
# Date:  19-Nov-2010
#
# Updates:
#     03-Feb-2012  Jdw  V3 PRD support --
#     31-Jan-2013  Jdw  Integrated within chemical reference data module
#     14-Jul-2013  jdw  add chemical depiction to report
#     14-Jun-2017  jdw  refactor change coordinate handling -
##
"""
PRD report generator -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import shutil
import logging

from mmcif_utils.chemcomp.PdbxChemCompIo import PdbxChemCompIo
from wwpdb.io.locator.ChemRefPathInfo import ChemRefPathInfo
from wwpdb.apps.chem_ref_data.report.ReportUtils import ReportUtils

logger = logging.getLogger(__name__)


class ChemCompReport(object):
    """PRD report generator functions."""

    def __init__(self, reqObj, verbose=False, log=sys.stderr):
        """PRD report generator.

        :param `verbose`:  boolean flag to activate verbose logging.
        :param `log`:      stream for logging.

        """
        self.__verbose = verbose
        self.__lfh = log
        #
        self.__reqObj = reqObj
        #
        self.__sObj = self.__reqObj.newSessionObj()
        self.__sessionPath = self.__sObj.getPath()
        self.__sessionRelativePath = self.__sObj.getRelativePath()
        self.__sessionId = self.__sObj.getId()
        #
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        self.__crPI = ChemRefPathInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)

        #
        self.__ccId = None
        self.__ccFilePath = None
        self.__ccFileFormat = "cif"
        #
        self.__rU = ReportUtils(verbose=self.__verbose, log=self.__lfh)
        #

    def setCcId(self, ccId):
        """Set an existing PRD identifier in archive collection as
        the report target
        """
        self.__ccId = str(ccId.upper())
        return self.__getFilePathFromId(self.__ccId)

    def setFilePath(self, ccFilePath, ccFileFormat="cif", ccId=None):
        self.__ccFilePath = ccFilePath
        self.__ccFileFormat = ccFileFormat
        if not os.access(self.__ccFilePath, os.R_OK):
            return False
        if ccId is not None:
            self.__ccId = str(ccId).upper()
        #
        return True

    def getFilePath(self):
        return self.__ccFilePath

    def __getFilePathFromId(self, ccId):
        """ """
        self.__ccFilePath = self.__crPI.getFilePath(ccId)
        if not os.access(self.__ccFilePath, os.R_OK):
            return False
        self.__ccFileFormat = "cif"
        return True

    #

    def doReport(self):
        """Return content required to render a chemical component tabular report."""
        #
        oD = {}
        oD["dataDict"] = {}
        filePath = self.__ccFilePath
        fileFormat = self.__ccFileFormat
        blockId = self.__ccId
        #
        logger.info("+ChemCompReport.doReport()  - Starting doReport()n")
        logger.info("+ChemCompReport.doReport()  - file path    %s", filePath)
        logger.info("+ChemCompReport.doReport()  - file format  %s", fileFormat)
        logger.info("+ChemCompReport.doReport()  - block Id     %s", blockId)

        #
        # make a local copy of the file (if required)
        #
        (_pth, fileName) = os.path.split(filePath)
        dirPath = os.path.join(self.__sessionPath, "report")
        localPath = os.path.join(dirPath, fileName)
        localRelativePath = os.path.join(self.__sessionRelativePath, "report", fileName)
        if filePath != localPath:
            if not os.access(dirPath, os.F_OK):
                os.makedirs(dirPath)
            shutil.copyfile(filePath, localPath)
            #

            logger.info("+ChemCompReport.doReport() - Copied input file %s to report session path %s", filePath, localPath)
        #
        # Path context --
        oD["idCode"] = self.__ccId
        oD["filePath"] = filePath
        oD["localPath"] = localPath
        oD["localRelativePath"] = localRelativePath
        oD["sessionId"] = self.__sessionId
        oD["editOpNumber"] = 0
        oD["requestHost"] = self.__reqObj.getValue("request_host")
        oD["xyzRelativePath"] = localRelativePath
        #
        imagePath = os.path.join(self.__sessionPath, self.__ccId + ".svg")
        ok = self.__rU.makeImagesFromPathList(pathList=[localPath], imagePath=imagePath, rows=1, cols=1, imageX=700, imageY=700)
        if ok:
            imageRelativePath = os.path.join(self.__sessionRelativePath, oD["idCode"] + ".svg")
            oD["imageRelativePath"] = imageRelativePath
        else:
            oD["imageRelativePath"] = None
        #
        try:
            ccR = PdbxChemCompIo(verbose=self.__verbose, log=self.__lfh)
            ccR.setFilePath(localPath, compId=blockId)

            oD["blockId"] = ccR.getCurrentContainerId()
            if self.__verbose:
                logger.info("+ChemCompReport.doReport() - category name list %r", ccR.getCurrentCategoryNameList())
            for catName in ccR.getCurrentCategoryNameList():
                oD["dataDict"][catName] = ccR.getCategory(catName=catName)
            hasExpt, hasIdeal = self.__rU.coordinatesExist(oD["dataDict"])
            oD["hasExpt"] = hasExpt
            oD["hasIdeal"] = hasIdeal
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failing with fileName %r", fileName)

        return oD



##
# File:  ChemRefReportDepictBootstrap.py
# Date:  18-Feb-2013  Jdw
#
# Updates:
#   2-Jun-2017 jdw add NGL viewer
#   3-Jun-2017 jdw fix CSS for NGL panes & missing
#  14-Jun-3017 jdw generalize the handling of coordinate files -
#                  change markup of tabbable section to toggle and resist jump scrolling -
#   Jun-2023 james smith add at-a-glance tab
#  27-Aug-2024  zf add "pdbx_chem_comp_pcm" category
##
"""
Create tabular HTML reports from chemical reference definitions.

This version uses Bootstrap CSS framework constructs.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import random
from wwpdb.apps.chem_ref_data.depict.ChemRefDataDepictBootstrap import (
    ChemRefDataDepictBootstrap,
)

import logging

logger = logging.getLogger(__name__)


class ChemRefReportDepictBootstrap(ChemRefDataDepictBootstrap):
    """Create tabular HTML reports from chemical reference definitions.

    This version uses Bootstrap CSS framework constructs.

    """

    def __init__(self, styleObject=None, verbose=False, log=sys.stderr):
        """

        :param `verbose`:  boolean flag to activate verbose logging.
        :param `log`:      stream for logging.

        """
        super(ChemRefReportDepictBootstrap, self).__init__(verbose, log)
        self.__debug = False
        #
        self.__st = styleObject
        #
        self.__setup()

    def __setup(self):
        """Category list --"""
        #
        if self.__st.getStyleId() in ["CHEM_COMP_V1"]:
            self.__reportCategories = [
                ("chem_comp", "chem_comp", "column-wise"),
                ("pdbx_chem_comp_synonyms", "synonyms", "row-wise"),
                ("chem_comp_atom", "chem_comp_atom", "row-wise"),
                ("chem_comp_bond", "chem_comp_bond", "row-wise"),
                ("pdbx_chem_comp_descriptor", "descriptor", "row-wise"),
                ("pdbx_chem_comp_identifier", "identifier", "row-wise"),
                ("pdbx_chem_comp_pcm", "pdbx_chem_comp_pcm", "row-wise"),
                ("pdbx_chem_comp_related", "related", "row-wise"),
                ("pdbx_chem_comp_atom_related", "atom_related", "row-wise"),
                ("pdbx_chem_comp_audit", "audit", "row-wise"),
            ]
        elif self.__st.getStyleId() in ["BIRD_V1"]:
            self.__reportCategories = [
                ("pdbx_reference_molecule_list", "molecule_list", "row-wise"),
                ("pdbx_reference_molecule_family", "molecule_family", "row-wise"),
                ("pdbx_reference_molecule_synonyms", "molecule_synonyms", "row-wise"),
                (
                    "pdbx_reference_molecule_annotation",
                    "molecule_annotation",
                    "row-wise",
                ),
                ("pdbx_reference_molecule_features", "molecule_features", "row-wise"),
                (
                    "pdbx_reference_molecule_related_structures",
                    "molecule_related_structures",
                    "row-wise",
                ),
                ("pdbx_reference_molecule_details", "details", "row-wise"),
                ("citation", "citation", "row-wise"),
                ("citation_author", "citation_author", "row-wise"),
                ("pdbx_family_prd_audit", "family_prd_audit", "row-wise"),
                #
                ("pdbx_reference_molecule", "molecule", "column-wise"),
                ("pdbx_reference_entity_list", "entity_list", "row-wise"),
                ("pdbx_reference_entity_src_nat", "entity_src_nat", "row-wise"),
                ("pdbx_reference_entity_poly", "entity_poly", "row-wise"),
                ("pdbx_reference_entity_poly_seq", "entity_poly_seq", "row-wise"),
                ("pdbx_reference_entity_sequence", "entity_sequence", "row-wise"),
                (
                    "pdbx_reference_entity_sequence_list",
                    "entity_sequence_list",
                    "row-wise",
                ),
                (
                    "pdbx_reference_entity_subcomponents",
                    "entity_subcomponents",
                    "row-wise",
                ),
                ("pdbx_reference_entity_nonpoly", "entity_nonpoly", "row-wise"),
                ("pdbx_reference_entity_link", "entity_link", "row-wise"),
                ("pdbx_reference_entity_poly_link", "entity_poly_link", "row-wise"),
                ("pdbx_prd_audit", "prd_audit", "row-wise"),
            ]

    def render(self, eD, style="tabs"):
        """ """
        if style in ["tabs"]:
            return self.__doRenderTabs(eD)
        elif style in ["accordion", "multiaccordion"]:
            return self.__doRenderAccordion(eD)
        else:
            return self.__doRenderPage(
                eD
            )  # Due to missing inheritance, and eD variable was missing - never come through here

    def __doRenderTabs(self, eD):
        """Render a tabbed table set.

        Bootstrap markup template  --

        <div class="tabbable"> <!-- Only required for left/right tabs -->
            <ul class="nav nav-tabs">
               <li class="active"><a href="#tab1" data-toggle="tab">Section 1</a></li>
               <li><a href="#tab2" data-toggle="tab">Section 2</a></li>
            </ul>

            <div class="tab-content">
               <div class="tab-pane active" id="tab1">
                   <p>I'm in Section 1.</p>
               </div>
                <div class="tab-pane" id="tab2">
                             <p>Howdy, I'm in Section 2.</p>
                </div>
             </div>
        </div>

        """

        idPrefix = "tab" + str(random.randint(0, 100000))
        catList = self.__reportCategories
        #
        if self.__debug:
            for ii, tup in enumerate(catList):
                logger.debug(
                    "ChemRefReportDepict (doRenderTabs) ii %d  tup %r", ii, tup
                )
            for ii, (x, y, z) in enumerate(catList):
                logger.debug(
                    "ChemRefReportDepict (doRenderTabs) ii %d  values  %s %s %s",
                    ii,
                    x,
                    y,
                    z,
                )
        #
        oL = []
        #
        # need for URL construction --
        cD = eD["dataDict"]
        idCode = eD["idCode"]
        #
        #
        # Write the tabs --
        #
        oL.append(
            '<div id="%s_report_section" class="tabbable results-section">' % idCode
        )
        # add close/dismiss --
        oL.append(
            '<button type="button" class="close" data-target="#%s_report_section" data-dismiss="alert" style="margin:7px; padding:1px;"> '
            % idCode
        )
        oL.append(
            '   <span aria-hidden="true"> <i class="fa fa-times"></i></span><span class="sr-only">Close</span></button>'
        )

        oL.append('<ul class="nav nav-tabs">')
        oL.append(
            '<li><a  class="active" data-target="#%s-tabs-id" data-toggle="tab">%s'
            % (idPrefix, idCode)
        )
        oL.append(
            '   <span aria-hidden="true"> <i class="fa fa-compress"></i></span><span class="sr-only">Close</span></a></li>'
        )
        iSection = 0
        for ii, (catName, catNameAbbrev, catStyle) in enumerate(catList):
            # For only popuated categories
            if catName in cD and (len(cD[catName]) > 0):
                oL.append(
                    '<li><a data-target="#%s-tabs-%d" data-toggle="tab">%s</a></li>'
                    % (idPrefix, ii, catNameAbbrev)
                )
                iSection = ii
        #
        if eD["imageRelativePath"] is not None:
            oL.append(
                '<li><a data-target="#%s-tabs-%d" data-toggle="tab">%s</a></li>'
                % (idPrefix, iSection + 1, "2D")
            )
        if eD["xyzRelativePath"] is not None:
            oL.append(
                '<li><a data-target="#%s-tabs-%d" data-toggle="tab" class="jsmol-section-%s">%s</a></li>'
                % (idPrefix, iSection + 2, idCode, "3D")
            )
        # at-a-glance for 3-or-5-letter cc id
        if (
            eD["imageRelativePath"] is not None
            and eD["xyzRelativePath"] is not None
            and (len(idCode) == 3 or len(idCode) == 5)
        ):
            oL.append(
                '<li><a data-target="#%s-tabs-%d" data-toggle="tab" class="ataglance-section-%s">%s</a></li>'
                % (idPrefix, iSection + 3, idCode, "at-a-glance")
            )

        oL.append("</ul>")
        #
        # Write the tables --
        #
        oL.append('<div  class="tab-content"> ')
        #
        oL.append('<div class="tab-pane active"  id="%s-tabs-id"></div>' % (idPrefix))

        iSection = 0
        for ii, (catName, catNameAbbrev, catStyle) in enumerate(catList):
            # For only popuated categories
            if catName in cD and (len(cD[catName]) > 0):
                oL.append('<div class="tab-pane"  id="%s-tabs-%d">' % (idPrefix, ii))
                oL.append(
                    '<table class="table table-striped table-bordered table-condensed" id="%s-%s">'
                    % (idPrefix, catName)
                )
                if catStyle == "column-wise":
                    self.__renderTableColumnWise(catName, cD[catName][0], oL)
                else:
                    self.__renderTableRowWise(catName, cD[catName], oL)

                oL.append("</table>")
                oL.append("</div>")
                iSection = ii
        #  2D image
        if eD["imageRelativePath"] is not None:
            oL.append(
                '<div class="tab-pane"  id="%s-tabs-%d">' % (idPrefix, iSection + 1)
            )
            oL.append(
                '<img src="%s" alt="%s" height="%d" width="%d">'
                % (eD["imageRelativePath"], idCode, 700, 700)
            )
            oL.append("</div>")

        # 3D app
        if eD["xyzRelativePath"] is not None:
            hasExpt = eD["hasExpt"]
            hasIdeal = eD["hasIdeal"]
            oL.append(
                '<div style="overflow:visible;" class="tab-pane jsmol-class-expt-%s jsmol-class-ideal-%s" data-payload="%s" id="%s-tabs-%d">'
                % (idCode, idCode, idCode, idPrefix, iSection + 2)
            )

            oL.append('<div style="display:flex;">')
            #             # h5 is ~15px + 20 vert margin
            if hasExpt:
                oL.append(
                    '  <div id="%s_jsmol_expt" style="display:inline-block; float:left; border: 2px solid lightgray; width:645px; height:645px; margin:2px; padding:1px; ">'
                    % idCode
                )
                oL.append("  </div>")
            if hasIdeal:
                oL.append(
                    '  <div id="%s_jsmol_ideal" style="display:inline-block; float:left; border: 2px solid lightgray; width:645px; height:645px; margin:2px; padding:1px; ">'
                    % idCode
                )
                oL.append("  </div>")

            oL.append("</div>")

            oL.append("</div>")

        # At-a-glance for 3-or-5-letter cc id
        if (
            eD["imageRelativePath"] is not None
            and eD["xyzRelativePath"] is not None
            and (len(idCode) == 3 or len(idCode) == 5)
        ):
            oL.append(
                '<div style="padding:10px;background-color:white;overflow:visible;" class="tab-pane tab-flex" id="%s-tabs-%d">'
                % (idPrefix, iSection + 3)
            )
            oL.append(
                '<table style="margin-right:20px;position:relative;display:inline-table;width:100ch;" id="%s-%s">'
                % (idPrefix, "chem_comp")
            )
            self.__renderTableAtAGlance("chem_comp", cD["chem_comp"][0], oL)
            oL.append("</table>")
            oL.append(
                '<img style="position:relative;display:inline-table;border:1px solid gray;" src="%s" alt="%s" height="%d" width="%d">'
                % (eD["imageRelativePath"], idCode, 500, 500)
            )
            oL.append(
                '<div id="%s_ataglance_ideal" style="position:relative;display:inline-table;border:1px solid lightgray;width:500px;height:500px;">'
                % idCode
            )
            oL.append("</div>")
            oL.append("</div>")
        #
        oL.append("</div>")
        oL.append("</div>")
        #
        return oL

    def __doRenderAccordion(self, eD):
        """
        Bootstrap accordion template  --

        <div class="accordion" id="accordion2">
            <div class="accordion-group">
               <div class="accordion-heading">
                 <a class="accordion-toggle" data-toggle="collapse" data-parent="#accordion2" href="#collapseOne">  Collapsible Group Item #1  </a>
               </div>
               <div id="collapseOne" class="accordion-body collapse in">
                   <div class="accordion-inner">
                         Anim pariatur cliche...
                   </div>
               </div>
             </div>

             <div class="accordion-group">
                 <div class="accordion-heading">
                     <a class="accordion-toggle" data-toggle="collapse" data-parent="#accordion2" href="#collapseTwo"> Collapsible Group Item #2 </a>
                 </div>
                  <div id="collapseTwo" class="accordion-body collapse">
                     <div class="accordion-inner">
                           Anim pariatur cliche...
                     </div>
                  </div>
             </div>
        </div>

        """
        #
        idPrefix = "acc" + str(random.randint(0, 100000))
        oL = []
        catList = self.__reportCategories
        cD = eD["dataDict"]
        #
        idTop = idPrefix + "-top"

        #
        # Write the tables --
        #
        oL.append('<div class="accordion" id="%s">' % idTop)

        #
        isFirst = True
        for ii, (catName, catNameAbbrev, catStyle) in enumerate(catList):
            # For only popuated categories
            if catName in cD and (len(cD[catName]) > 0):
                active = "in" if isFirst else ""
                idSection = idPrefix + "-sec-" + str(ii)
                oL.append('<div class="accordion-group">')
                oL.append('<div class="accordion-heading">')
                oL.append(
                    '<a class="accordion-toggle" data-toggle="collapse" data-parent="#%s" href="#%s">%s</a>'
                    % (idTop, idSection, catNameAbbrev)
                )
                oL.append("</div>")
                oL.append(
                    '<div id="%s" class="accordion-body collapse %s">'
                    % (idSection, active)
                )
                oL.append('<div  class="accordion-inner">')
                #
                oL.append(
                    '<table class="table table-striped table-bordered table-condensed">'
                )
                if catStyle == "column-wise":
                    self.__renderTableColumnWise(catName, cD[catName][0], oL)
                else:
                    self.__renderTableRowWise(catName, cD[catName], oL)
                oL.append("</table>")
                #
                oL.append("</div>")
                oL.append("</div>")
                #
                oL.append("</div> <!-- end of accordion group -->")
        oL.append("</div> <!-- end of accordion -->")
        #
        return oL

    def __doRenderPage(self, eD):
        """Render a full page report with home brewed multi-accordion presentation style"""
        oL = []
        oL.append(self._pragmaXhtml)  # pylint: disable=no-member
        oL.append("<html>")
        oL.append("<head>")
        oL.append(self._jQueryGenericInclude)  # pylint: disable=no-member

        # oL.append(self._jQueryTableInclude)
        oL.append("</head>")

        oL.append('<body class="oneColLiqHdr">')
        oL.append(self._menuCommonIncludeInline)  # pylint: disable=no-member

        tableList = [t[0] for t in self.__reportCategories]
        self.__jQueryReportScript1(tableList, eD, oL)

        oL.append("<h2>Peptide Reference Dictionary Report</h2>")
        oL.append("<br/>")

        #
        cD = eD["dataDict"]

        for catName in ["chem_comp"]:
            if catName in cD and (len(cD[catName]) > 0):
                oL.append('<div class="sectionbar1">')
                oL.append(
                    '  <a class="sectionbar1" href="" id="toggle_section_%s">Show</a> Category: %s'
                    % (catName, catName)
                )
                oL.append("</div>")

                oL.append(
                    '<div style="display: block;" id="d_%s" class="displaynone">'
                    % catName
                )
                oL.append('<table id="%s">' % catName)
                self.__renderTableColumnWise(catName, cD[catName][0], oL)
                oL.append("</table>")
                oL.append("</div>")
        #
        # <div class="sb0"><a class="sb0" href="" id="toggle_section_XXXX">Show</a></div> <div class="sb1">Category:  Entity Reference</div>
        #
        for catName in tableList:
            if catName in cD and (len(cD[catName]) > 0):
                oL.append('<div class="sb0">')
                oL.append(
                    '  <a class="sb0" href="" id="toggle_section_%s">Show</a>' % catName
                )
                oL.append("</div>")
                oL.append('<div class="sb1">Category: %s </div>' % catName)
                oL.append("<br />")

                oL.append(
                    '<div style="display: block;" id="d_%s" class="displaynone">'
                    % catName
                )
                oL.append('<table id="%s">' % catName)
                if catName in ["pdbx_reference_molecule"]:
                    self.__renderTableColumnWise(catName, cD[catName][0], oL)
                else:
                    self.__renderTableRowWise(catName, cD[catName], oL)
                oL.append("</table>")
                oL.append("</div>")

        oL.append(self._trailer)  # pylint: disable=no-member
        oL.append("</body>")
        oL.append("</html>")
        #
        return oL

    def __renderTableColumnWise(self, catName, rD, oL):
        """Render table with unit cardinality.  Columns for the single row are listed vertically
        to the left of column values.
        """

        #
        iCol = 0
        self.__markupRow(catName, rD)
        #
        for itemName, itemDefault in self.__st.getItemNameAndDefaultList(catName):
            if itemName in rD:
                itemValue = rD[itemName]
            else:
                itemValue = itemDefault

            oL.append("<tr>")
            oL.append("<td>%s</td>" % self.__attributePart(itemName))

            oL.append("<td>%s</td>" % (itemValue))
            oL.append("</tr>")
            iCol += 1

    def __renderTableRowWise(self, catName, rL, oL):
        """Render a multirow table."""
        # Column labels --
        oL.append("<tr>")
        for itemName in self.__st.getItemNameList(catName):
            oL.append("<th>%s</th>" % self.__attributePart(itemName))
        oL.append("</tr>")
        #
        # Column data ---
        #
        iRow = 0
        for row in rL:
            self.__markupRow(catName, row)
            self.__renderRow(catName, row, iRow, oL, insertDefault=False)
            iRow += 1
        #
        #

    def __renderTableAtAGlance(self, catName, rD, oL):
        """One-column rendering"""
        #
        # iCol = 0
        self.__markupRow(catName, rD)
        #
        lst = self.__st.getItemNameAndDefaultList(catName)
        d = {k: v for k, v in lst}
        # reorder keys for at-a-glance
        keys = [
            ("_chem_comp.id", "ID"),
            ("_chem_comp.pdbx_release_status", "Status"),
            ("_chem_comp.name", "Name"),
            ("_chem_comp.pdbx_synonyms", "Synonyms"),
            ("_chem_comp.formula", "Formula"),
            ("_chem_comp.formula_weight", "Formula weight"),
            ("_chem_comp.pdbx_formal_charge", "Formal charge"),
            ("_chem_comp.type", "Type"),
            ("_chem_comp.pdbx_type", "Pdbx_type"),
            ("_chem_comp.mon_nstd_parent_comp_id", "Parent"),
            ("_chem_comp.pdbx_subcomponent_list", "Subcomponents"),
            ("_chem_comp.pdbx_replaces", "Replace"),
            ("_chem_comp.pdbx_replaced_by", "Replace by"),
            (
                "_chem_comp.pdbx_model_coordinates_missing_flag",
                "Model coordinates missing",
            ),
            ("_chem_comp.pdbx_model_coordinates_db_code", "Model PDB code"),
            ("_chem_comp.pdbx_initial_date", "Initial date"),
            ("_chem_comp.pdbx_modified_date", "Modified date"),
            ("_chem_comp.pdbx_processing_site", "Site"),
        ]
        for k, v in keys:
            oL.append("<tr>")
            oL.append("<td>%s:&nbsp%s</td>" % (v, rD[k] if k in rD else d[k]))
            oL.append("</tr>")

    def __renderRow(
        self, catName, row, iRow, oL, insertDefault=False
    ):  # pylint: disable=unused-argument
        """Render a row in a multirow table."""
        oL.append("<tr>")
        #
        for itemName, itemDefault in self.__st.getItemNameAndDefaultList(catName):
            if insertDefault:
                itemValue = itemDefault
            elif itemName in row:
                itemValue = row[itemName]
            else:
                itemValue = itemDefault

            oL.append("<td>%s</td>" % (itemValue))
        oL.append("</tr>")
        #

    def __markupLinks(self, cName, cVal):
        """Add markup to selected column types -"""
        #
        rst = cVal
        try:
            if cName in ["family", "prd"]:
                rst = '<a class="app-ref-report"  href="#">%s</a>' % cVal
            elif cName in ["cc"]:
                rpt = '<a class="app-ref-report"  href="#">%s</a>' % cVal
                le = (
                    '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>'
                    % cVal
                )
                rst = "%s &nbsp; %s" % (rpt, le)
            elif cName in ["id"]:
                if cVal.startswith("PRD_") or cVal.startswith("FAM_"):
                    rst = '<a class="app-ref-report"  href="#">%s</a>' % cVal
                else:
                    rpt = '<a class="app-ref-report"  href="#">%s</a>' % cVal
                    le = (
                        '<a target="_blank" href="http://ligand-expo.rcsb.org/pyapps/ldHandler.py?formid=cc-index-search&operation=ccid&target=%s">(LE)</a>'
                        % cVal
                    )
                    rst = "%s &nbsp; %s" % (rpt, le)
        except Exception as e:
            logger.info(
                "ChemRefReportDepict (markuplinks) failing cName %r cVal %r %r",
                cName,
                cVal,
                str(e),
            )
            logger.exception("Failure in __markupLinks")
        return rst

    def __markupRow(self, catName, rD):
        """Markup a row (row dictionary) in the input category."""
        if catName == "pdbx_reference_molecule_list":
            itemName = "_pdbx_reference_molecule_list.prd_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) > 5 and itemValue.startswith("PRD_"):
                    rD[itemName] = self.__markupLinks("prd", itemValue)
        if catName == "pdbx_reference_molecule_synonyms":
            itemName = "_pdbx_reference_molecule_synonyms.prd_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) > 5 and itemValue.startswith("PRD_"):
                    rD[itemName] = self.__markupLinks("prd", itemValue)
            itemName = "_pdbx_reference_molecule_synonyms.chem_comp_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)
        if catName == "pdbx_reference_molecule_features":
            itemName = "_pdbx_reference_molecule_features.prd_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) > 5 and itemValue.startswith("PRD_"):
                    rD[itemName] = self.__markupLinks("prd", itemValue)
            itemName = "_pdbx_reference_molecule_features.chem_comp_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)
        if catName == "pdbx_reference_molecule_details":
            itemName1 = "_pdbx_reference_molecule_details.source"
            itemName2 = "_pdbx_reference_molecule_details.source_id"
            if itemName1 in rD:
                srcType = str(rD[itemName1]).upper()
                srcValue = rD[itemName2]
                if srcType == "DOI" and len(srcValue) > 2:
                    rD[
                        itemName2
                    ] = '<a target="_blank" href="http://dx.doi.org/%s">%s</a>' % (
                        srcValue,
                        srcValue,
                    )
                if srcType == "PUBMED" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=search&db=pubmed&term=%s">%s</a>'
                        % (srcValue, srcValue)
                    )
                if srcType == "DRUGBANK" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://www.drugbank.ca/cgi-bin/getCard.cgi?CARD=%s">%s</a>'
                        % (srcValue, srcValue)
                    )
                if srcType == "PUBCHEM" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://pubchem.ncbi.nlm.nih.gov/summary/summary.cgi?cid=%s">%s</a>'
                        % (srcValue, srcValue)
                    )

                if srcType == "URL" and len(srcValue) > 2:
                    rD[itemName2] = '<a target="_blank" href="%s">%s</a>' % (
                        srcValue,
                        srcValue,
                    )

                if srcType == "PMCID" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://www.ncbi.nlm.nih.gov/pmc/?term=%s">%s</a>'
                        % (srcValue, srcValue)
                    )

                if srcType == "PMC" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://www.ncbi.nlm.nih.gov/pmc/?term=%s">%s</a>'
                        % (srcValue, srcValue)
                    )

                if srcType == "UNIPROT" and len(srcValue) > 2:
                    rD[itemName2] = (
                        '<a target="_blank" href="http://www.uniprot.org/uniprot/%s">%s</a>'
                        % (srcValue, srcValue)
                    )

        if catName == "citation":
            itemName = "_citation.pdbx_database_id_DOI"
            if itemName in rD and len(rD[itemName]) > 1:
                itemValue = rD[itemName]
                rD[
                    itemName
                ] = '<a target="_blank" href="http://dx.doi.org/%s">%s</a>' % (
                    itemValue,
                    itemValue,
                )
            itemName = "_citation.pdbx_database_id_PubMed"
            if itemName in rD and len(rD[itemName]) > 1:
                itemValue = rD[itemName]
                rD[itemName] = (
                    '<a target="_blank" href="http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=search&db=pubmed&term=%s">%s</a>'
                    % (itemValue, itemValue)
                )

        if catName == "pdbx_reference_molecule":
            itemName = "_pdbx_reference_molecule.representative_PDB_id_code"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) > 3:
                    rD[itemName] = (
                        '<a target="_blank" href="http://www.rcsb.org/pdb/explore/explore.do?structureId=%s">%s</a>'
                        % (itemValue, itemValue)
                    )

            itemName = "_pdbx_reference_molecule.chem_comp_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)
        if catName == "pdbx_reference_entity_nonpoly":
            itemName = "_pdbx_reference_entity_nonpoly.chem_comp_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)
        if catName == "pdbx_reference_entity_poly_seq":
            itemName = "_pdbx_reference_entity_poly_seq.mon_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)
            itemName = "_pdbx_reference_entity_poly_seq.parent_mon_id"
            if itemName in rD:
                itemValue = rD[itemName]
                if len(itemValue) >= 3:
                    rD[itemName] = self.__markupLinks("cc", itemValue)

    # ------

    def __jQueryReportScript1(self, tableList, eD, oL):
        # Context that will get encoded for call back --
        filePath = eD["filePath"]
        localPath = eD["localPath"]
        localRelativePath = eD["localRelativePath"]
        sessionId = eD["sessionId"]
        editOpNumber = eD["editOpNumber"]
        blockId = eD["blockId"]
        #
        #
        # Report controls
        #
        oL.append('<script type="text/javascript">')
        oL.append("     var editNumber        = %d;" % editOpNumber)
        oL.append('     var filePath          = "%s";' % filePath)
        oL.append('     var localPath         = "%s";' % localPath)
        oL.append('     var localRelativePath = "%s";' % localRelativePath)
        oL.append('     var blockId           = "%s";' % blockId)
        oL.append('     var sessionId         = "%s";' % sessionId)
        oL.append("     var ajaxTimeout       = 60000;")
        oL.append("$(document).ready(function(){")
        #
        oL.append("    $.ajaxSetup({")
        oL.append('        type: "POST",')
        oL.append('        dataType: "json",')
        oL.append("        async: true,")
        oL.append("        timeout: ajaxTimeout,")
        oL.append("        cache: false")
        oL.append("    });")
        #
        #
        # oL.append('	$("#%s").tablesorter();' % tableId)
        #
        #
        dS = ""
        for t in tableList[:-1]:
            dS += "#d_%s, " % t
        dS += "#d_%s" % tableList[-1]
        oL.append('$("%s").hide();' % dS)

        #
        tS = ""
        for t in tableList[:-1]:
            tS += "#toggle_section_%s, " % t
        tS += "#toggle_section_%s" % tableList[-1]

        oL.append('$("%s").click(function() {' % tS)
        oL.append('        myConsoleLog("TOGGLE VALUE "+$(this).text());')
        # oL.append('        $(this).parents("div").filter(":first").next().toggle(400);')
        oL.append(
            '        $(this).parents("div").filter(":first").next().next().next().toggle(400);'
        )
        oL.append('        $(this).text($(this).text() == "Show" ? "Hide" : "Show");')
        oL.append("        return false;")
        oL.append(" });")

        oL.append(' $(".sb0").corner("round");')
        oL.append(' $(".sb1").corner("round");')
        #
        # button controls --
        # self.__jQueryButtonScripts(eD,oL)
        #
        oL.append("});")

        oL.append("</script>")

    def __attributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1 :]



##
# File:  ChemRefDataWebApp.py
# Date:  5-Nov-2010
#
# Updates:
#  6-Nov-2010 Add rdbms load and cvs sync
#
#  3-Feb-2012 jdw Support for V3 PRD and family files
# 12-Sep-2012 jdw add CVS add and update for PRDCC files --
#
# 25-Nov-2012 jdw migrate to wwPDB new framework
# 24-Jan-2013 jdw add id search options
# 25-Jan-2013 jdw replace scripted cvs sync and dbload with new api calls.
# 28-Jan-2013 jdw replace scripted update/add with new api calls.
# 30-Jan-2013 jdw integrate into chemical reference data package.
#  2-Feb-2013 jdw revise return values for AJAX methods.
#  4-Feb-2013 jdw Add classes ChemRefDataCvsUtils() & ChemRefDataDbUtils() which
#                 support all chemical reference data.
#  4-Feb-2013 jdw Add service methods for chemical component data.
#  *-Feb-2013 jdw Much refactoring for Bootstrap framework.
# 20-Feb-2013 jdw Use common WebRequest module.
# 23-May-2017 jdw Overhaul - strip all old methods
#
##
"""
Chemical Reference Data (CRD) tool web request and response processing modules.

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import sys
import shutil
import types
import ntpath

from wwpdb.io.file.mmCIFUtil import mmCIFUtil
from wwpdb.utils.session.WebRequest import InputRequest, ResponseContent

# from wwpdb.utils.session.UtilDataStore import UtilDataStore
#
from wwpdb.apps.chem_ref_data.report.BirdReport import BirdReport
from wwpdb.apps.chem_ref_data.report.ChemCompReport import ChemCompReport
from wwpdb.apps.chem_ref_data.report.ChemRefReportDepictBootstrap import ChemRefReportDepictBootstrap

#
from wwpdb.apps.chem_ref_data.search.ChemRefSearch import ChemRefSearch
from wwpdb.apps.chem_ref_data.search.ChemRefSearchDepictBootstrap import ChemRefSearchDepictBootstrap

from wwpdb.apps.chem_ref_data.utils.ChemRefDataCvsUtils import ChemRefDataCvsUtils
from wwpdb.apps.chem_ref_data.utils.ChemRefDataDbUtils import ChemRefDataDbUtils
from wwpdb.io.locator.ChemRefPathInfo import ChemRefPathInfo
from wwpdb.apps.chem_ref_data.utils.DownloadUtils import DownloadUtils
from wwpdb.apps.chem_ref_data.utils.ChemRefDataMiscUtils import ChemRefDataMiscUtils

from mmcif_utils.style.PrdCategoryStyle import PrdCategoryStyle
from mmcif_utils.style.ChemCompCategoryStyle import ChemCompCategoryStyle

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCommon
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility

#
import logging

logger = logging.getLogger(__name__)

#


class ChemRefDataWebApp(object):
    """Handle request and response object processing for the chemical reference data services."""

    def __init__(self, parameterDict=None, verbose=False, log=sys.stderr, siteId="WWPDB_DEPLOY_TEST"):
        """
        Create an instance of `ChemRefDataWebApp` to manage a chemical reference data  web requests.

         :param `parameterDict`: dictionary storing parameter information from the web request.
             Storage model for GET and POST parameter data is a dictionary of lists.
         :param `verbose`:  boolean flag to activate verbose logging.
         :param `log`:      stream for logging.

        """
        if parameterDict is None:
            parameterDict = {}

        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False
        self.__siteId = siteId
        #
        self.__cICommon = ConfigInfoAppCommon(self.__siteId)
        self.__topPath = self.__cICommon.get_site_web_apps_top_path()
        self.__topSessionPath = self.__cICommon.get_site_web_apps_top_sessions_path()
        #

        if isinstance(parameterDict, dict):
            self.__myParameterDict = parameterDict
        else:
            self.__myParameterDict = {}

        logger.info("+ChemRefDataWebApp.__init() - STARTING REQUEST ------------------------------------")

        self.__reqObj = InputRequest(self.__myParameterDict, verbose=self.__verbose, log=self.__lfh)

        self.__templatePath = os.path.join(self.__topPath, "htdocs", "chem_ref_data_ui")
        #
        self.__reqObj.setValue("TopSessionPath", self.__topSessionPath)
        self.__reqObj.setValue("TemplatePath", self.__templatePath)
        self.__reqObj.setValue("AppHtdocsPath", self.__templatePath)
        self.__reqObj.setValue("TopPath", self.__topPath)
        self.__reqObj.setValue("WWPDB_SITE_ID", self.__siteId)
        os.environ["WWPDB_SITE_ID"] = self.__siteId

        #
        self.__reqObj.setDefaultReturnFormat(return_format="html")
        #
        if self.__debug:
            self.__reqObj.printIt(ofh=self.__lfh)
            logger.info("+ChemRefDataWebApp.__init() - completed\n---------------------------------------------\n")
            self.__lfh.flush()

    def doOp(self):
        """Execute request and package results in a response dictionary object.

        :returns:
             A dictionary containing response data for the input request.
             Minimally, the content of this dictionary will include the
             keys: CONTENT_TYPE and REQUEST_STRING.

        """
        stw = ChemRefDataWebAppWorker(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        rC = stw.doOp()
        if self.__debug:
            rqp = self.__reqObj.getRequestPath()
            logger.info("+ChemRefDataWebApp.doOp() request path %s", rqp)
            logger.info("+ChemRefDataWebApp.doOp() return format %s", self.__reqObj.getReturnFormat())
            if rC is not None:
                logger.debug("%s", ("".join(rC.dump())))
            else:
                logger.info("+ChemRefDataWebApp.doOp() return object is empty")
        #
        return rC.get()


class ChemRefDataWebAppWorker(object):
    def __init__(self, reqObj=None, verbose=False, log=sys.stderr):
        """
        Worker methods for the chemical reference data module.

        Performs URL -> application mapping for this module.

        All operations can be driven from this interface which can
        supplied with control information from web application request
        or from a testing application.


        """
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__reqObj = reqObj
        self.__sObj = None
        self.__sessionId = None
        self.__sessionPath = None
        self.__rltvSessionPath = None

        #
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        self.__cI = ConfigInfo(self.__siteId)
        # self.__cICommon = ConfigInfoAppCommon(self.__siteId)
        self.__crPI = ChemRefPathInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)

        #
        # self.__uds = UtilDataStore(reqObj=self.__reqObj, prefix=None, verbose=self.__verbose, log=self.__lfh)

        #
        # Reference data configuration items include:
        #
        #
        self.__appPathD = {
            "/service/environment/dump": "_dumpOp",
            # -------------------
            "/service/chemref/search": "_chemRefFullSearchOp",
            "/service/chemref/search/autocomplete": "_chemRefFullSearchAutoCompeteOp",
            # -------------------
            "/service/chemref/newsession": "_newSessionOp",
            # '/service/chemref/getsessioninfo':                    '_getSessionInfoOp',
            "/service/chemref/adminops": "_chemRefAdminOps",
            "/service/chemref/inline_idops": "_chemRefInlineIdOps",
            "/service/chemref/inline_fileops": "_chemRefInlineFileOps",
            "/service/chemref/editor": "_chemRefEditorOps",
        }

    def doOp(self):
        """Map operation to path and invoke operation.  Exceptions are caught within this method.

        :returns:

        Operation output is packaged in a ResponseContent() object.

        """
        #
        try:
            reqPath = self.__reqObj.getRequestPath()
            if reqPath not in self.__appPathD:
                # bail out if operation is unknown -
                rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                rC.setError(errMsg="Unknown operation")
            else:
                mth = getattr(self, self.__appPathD[reqPath], None)
                rC = mth()
            return rC
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("doOp failing")
            rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            rC.setError(errMsg="Operation failure")
            return rC

    # ------------------------------------------------------------------------------------------------------------
    #      Top-level REST methods
    #
    #
    def _dumpOp(self):
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        rC.setHtmlList(self.__reqObj.dump(format="html"))
        return rC

    #
    # ---------------------------------------------------------------------------------------------------
    #
    #                      File-based methods on the ADMIN path implemented with JSON responses -
    #
    def _chemRefInlineFileOps(self):
        """Chemical reference data operations on uploaded files."""
        operation = self.__reqObj.getValue("operation")
        logger.info("+ChemRefDataWebAppWorker._chemRefInlineFileOps() starting with op %s", operation)

        isFile = False
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)

        filePath = None
        rootName = None
        if self.__isFileUpload():
            # make a copy of the file in the session directory and set 'fileName'
            self.__uploadFile()
            fileName = self.__reqObj.getValue("fileName")
            filePath = os.path.join(self.__sessionPath, fileName)
            (rootName, _ext) = os.path.splitext(fileName)
            isFile = True

        logger.info("+ChemRefDataWebAppWorker._chemRefInlineFileOps() filePath %s", filePath)

        #
        aTagList = []
        htmlList = []
        webPathList = []
        myIdList = []
        hasDiags = False

        if not isFile or fileName is None or len(fileName) < 1:
            rC.setError(errMsg="File upload failed.")

        elif operation in ["report"]:
            du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            du.fetchFile(filePath)
            idCode = du.getIdFromFileName(filePath)
            # myIdList.append(idCode)
            repType = du.getIdType(idCode)
            downloadPath = du.getDownloadPath()
            # webPathList.append(du.getWebPath())
            aTagList.append(du.getAnchorTag())
            #
            oL, _, webXyzPath = self.__makeTabularReport(filePath=downloadPath, repositoryType=repType, idCode=idCode, layout="tabs")
            htmlList.extend(oL)
            myIdList.append(idCode)
            webPathList.append(webXyzPath)

            if len(aTagList) > 0:
                rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;&nbsp;".join(aTagList))
                rC.setHtmlList(htmlList)
                rC.setStatus(statusMsg="Reports completed")
                rC.set("webPathList", webPathList)
                rC.set("idCodeList", myIdList)
            else:
                rC.setError(errMsg="Report preparation failed")

        elif operation in ["check", "cvsupdate", "cvsadd"]:
            logPath = os.path.join(self.__sessionPath, rootName + "-cif-check.log")
            self.__removeFile(logPath)
            hasDiags = self.__makeCifCheckReport(filePath, logPath)
            if hasDiags:
                du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                du.fetchFile(logPath)
                aTagList.append(du.getAnchorTag())
                rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;&nbsp;".join(aTagList))
                rC.setStatus(statusMsg="Check completed")
            else:
                rC.setHtmlLinkText("")
                rC.setStatus(statusMsg="No diagnostics for %s" % fileName)

        if hasDiags:
            return rC

        if operation == "cvsupdate":
            logPath = os.path.join(self.__sessionPath, rootName + "-cvs-update.log")
            self.__removeFile(logPath)
            ok = self.__doCvsUpdate(filePath, logPath)
            if ok:
                du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                du.fetchFile(logPath)
                aTagList.append(du.getAnchorTag())
                rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))
                rC.setStatus(statusMsg="Update completed")
            else:
                rC.setStatus(statusMsg="Update failed for %s" % fileName)
        elif operation == "cvsadd":
            logPath = os.path.join(self.__sessionPath, rootName + "-cvs-add.log")
            self.__removeFile(logPath)
            ok = self.__doCvsAdd(filePath, logPath)
            if ok:
                du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                du.fetchFile(logPath)
                aTagList.append(du.getAnchorTag())
                rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))
                rC.setStatus(statusMsg="Add completed")
            else:
                rC.setStatus(statusMsg="Add failed for %s" % fileName)
        elif operation == "annotatecomp":
            logPath = os.path.join(self.__sessionPath, rootName + "-annot-comp.log")
            self.__removeFile(logPath)
            outPath = os.path.join(self.__sessionPath, rootName + "-annot.cif")
            self.__removeFile(outPath)
            ok = self.__doAnnotateChemComp(filePath, outPath, logPath)
            #
            if ok:
                du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                ok1 = du.fetchFile(outPath)
                if ok1:
                    aTagList.append(du.getAnchorTag())
                ok1 = du.fetchFile(logPath)
                if ok1:
                    aTagList.append(du.getAnchorTag())

                rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))
                rC.setStatus(statusMsg="Operation completed")
            else:
                rC.setStatus(statusMsg="Operation failed for %s" % fileName)
        else:
            pass

        return rC

    def __removeFile(self, filePath):
        try:
            os.remove(filePath)
        except:  # noqa: E722 pylint: disable=bare-except
            return False
        return True

    def __doCvsUpdate(self, filePath, logPath):
        """ """
        if self.__verbose:
            logger.info("+ChemRefDataWebAppWorker._doCvsUpdate() starting for %s", filePath)

        try:
            crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok, textList = crdu.updateFile(filePath)
            ofh = open(logPath, "w")
            ofh.write("%s" % "\n".join(textList))
            ofh.close()
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("cvs update failing")
            return False

    def __doCvsAdd(self, filePath, logPath):
        """ """
        if self.__verbose:
            logger.info("+ChemRefDataWebAppWorker._doCvsAdd() starting for %s", filePath)

        try:
            crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok, textList = crdu.addFile(filePath)
            ofh = open(logPath, "w")
            ofh.write("%s" % "\n".join(textList))
            ofh.close()
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("cvs add failing")
            return False

    def __doCvsRemove(self, idCode, logPath):
        """ """
        if self.__verbose:
            logger.info("+ChemRefDataWebAppWorker._doCvsRemove() starting for %s", idCode)
        try:
            crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok, textList = crdu.remove(idCode)
            ofh = open(logPath, "w")
            ofh.write("%s" % "\n".join(textList))
            ofh.close()
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("cvs remove failing")
            return False

    def __doCvsHistory(self, idCode, logPath):
        """ """
        if self.__verbose:
            logger.info("+ChemRefDataWebAppWorker._doCvsHistory() starting for %s", idCode)

        try:
            crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok, textList = crdu.history(idCode)
            ofh = open(logPath, "w")
            ofh.write("%s" % "\n".join(textList))
            ofh.close()
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("cvs history failing")
            return False

    def __doCvsCheckoutRevisions(self, idCode, pathList):
        """ """
        if self.__verbose:
            logger.info("+ChemRefDataWebAppWorker._doCvsCheckOutRevisions() starting for %s", idCode)

        try:
            crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok, pL = crdu.checkoutRevisions(idCode)
            pathList.extend(pL)
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("cvs checkout revisions failing")
            return False

    def __makeCifCheckReport(self, filePath, logPath):
        """Create CIF dictionary check on the input file and return diagnostics in logPath.

        Return True if a report is created (logPath exists and has non-zero size)
            or False otherwise
        """
        logger.info("+ChemRefDataWebAppWorker.__makeCifCheckReport() for file %s", filePath)
        logger.info("+ChemRefDataWebAppWorker._cifCheckOp() uploaded file %s", filePath)

        dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
        dp.imp(filePath)
        dp.op("check-cif")
        dp.exp(logPath)
        dp.cleanup()
        if os.access(logPath, os.R_OK) and os.stat(logPath).st_size > 0:
            return True
        else:
            return False

    def __doAnnotateChemComp(self, inpPath, outPath, logPath):
        """Add annotations to the input chemical component definition and return the annotated file and diagnostics in logPath.

        Return True if the operation completes (outPath exists and has non-zero size)
            or False otherwise
        """
        logger.info("Starting for chemical component file %s", inpPath)
        #
        dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
        dp.setDebugMode(flag=True)
        dp.imp(inpPath)
        dp.op("chem-comp-annotate-comp")
        dp.exp(outPath)
        dp.expLog(logPath)
        # dp.cleanup()
        if os.access(outPath, os.R_OK) and os.stat(outPath).st_size > 0:
            return True
        else:
            return False

    # ----------------------------------------------------------------------------------------------------
    #             ID-based methods on the ADMIN path implemented with JSON responses -
    #             JSON ID-level operations -
    #
    def _chemRefInlineIdOps(self):
        """Chemical reference data operations on id codes."""
        operation = self.__reqObj.getValue("operation")
        logger.info("+ChemRefDataWebAppWorker._chemRefInlineIdOps() starting with op %s", operation)

        idCodes = self.__reqObj.getValue("idcode")
        idCodeList = idCodes.split(" ")

        logger.info("+ChemRefDataWebAppWorker._chemRefInlineIdOps() fetch id(s) %r", idCodeList)
        #
        if operation == "fetch":
            return self.__makeIdListFetchResponse(idCodeList)
        elif operation == "report":
            return self.__makeIdListReportResponse(idCodeList)
        elif operation == "history":
            return self.__makeIdListCvsHistoryResponse(idCodeList)
        elif operation == "revfetch":
            return self.__makeRevisionCheckoutResponse(idCodeList)
        elif operation == "remove":
            return self.__makeIdListCvsRemoveResponse(idCodeList)

        else:
            pass

    def __makeIdListFetchResponse(self, idCodeList):
        """ """
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        aTagList = []
        for idCode in idCodeList:
            filePath = self.__crPI.getFilePath(idCode=idCode)
            if du.fetchFile(filePath):
                aTagList.append(du.getAnchorTag())

        if len(aTagList) > 0:
            rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))
            rC.setStatus(statusMsg="Fetch completed")
        else:
            rC.setError(errMsg="No corresponding reference file(s)")
            # do nothing

        return rC

    def __makeIdListCvsHistoryResponse(self, idCodeList):
        """ """
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        aTagList = []
        #
        success = True
        du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        for idCode in idCodeList:
            logPath = os.path.join(self.__sessionPath, idCode + "-cvs-history.log")
            ok = self.__doCvsHistory(idCode, logPath)
            if not ok:
                success = False
            ok1 = du.fetchFile(logPath)
            if ok1:
                aTagList.append(du.getAnchorTag())

        if len(aTagList) > 0:
            rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))

        if success:
            rC.setStatus(statusMsg="History operations completed")
        else:
            rC.setError(errMsg="History operations failed")

        return rC

    def __makeRevisionCheckoutResponse(self, idCodeList):
        """ """
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        aTagList = []
        pathList = []
        #
        success = False
        du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        for idCode in idCodeList:
            ok = self.__doCvsCheckoutRevisions(idCode, pathList)
            if ok:
                for pth in pathList:
                    ok1 = du.fetchFile(pth)
                    if ok1:
                        aTagList.append(du.getAnchorTag())

        if len(aTagList) > 0:
            success = True
            rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))

        if success:
            rC.setStatus(statusMsg="Revision checkout completed")
        else:
            rC.setError(errMsg="Revision checkout failed")

        return rC

    def __makeIdListCvsRemoveResponse(self, idCodeList):
        """ """
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        errorMessage = ""
        for idCode in idCodeList:
            if len(idCode) > 3:
                continue
            #
            filePath = self.__crPI.getFilePath(idCode=idCode)
            if (not filePath) or (not os.access(filePath, os.R_OK)):
                errorMessage += "Invalid CC ID: " + idCode + "\n"
                continue
            #
            try:
                cifObj = mmCIFUtil(filePath=filePath)
                relStatus = cifObj.GetSingleValue("chem_comp", "pdbx_release_status").strip().upper()
                if relStatus == "REL":
                    errorMessage += "The ligand '" + idCode + "' has REL status and it should not be removed.\n"
                #
            except:  # noqa: E722 pylint: disable=bare-except
                logger.exception("Failure in __makeIdListCvsRemoveResponse")
            #
        #
        if errorMessage:
            rC.setError(errMsg=errorMessage)
            return rC
        #
        aTagList = []
        #
        success = True
        du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        for idCode in idCodeList:
            logPath = os.path.join(self.__sessionPath, idCode + "-cvs-remove.log")
            ok = self.__doCvsRemove(idCode, logPath)
            if not ok:
                success = False
            ok1 = du.fetchFile(logPath)
            if ok1:
                aTagList.append(du.getAnchorTag())

        if len(aTagList) > 0:
            rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;nbsp;".join(aTagList))

        if success:
            rC.setStatus(statusMsg="Remove operations completed")
        else:
            rC.setError(errMsg="Remove operations failed")

        return rC

    #  ---------------------------------------------------------------------------------------------
    #
    #     Current search methods - searchType = <search_attributes, ...>|<query_type>
    ##
    def __getSearchType(self, searchTypeInput):
        """Return: searchType,queryType, target input type, comparison type"""
        try:
            tL = str(searchTypeInput).split("|")
            if len(tL) > 1:
                return tL[0], tL[1], tL[2], tL[3]
            else:
                return tL[0], None, None, None
        except Exception as e:
            logger.exception("Failing with input %r and %r", searchTypeInput, str(e))

        return None, None, None, None

    def _chemRefFullSearchAutoCompeteOp(self):
        self.__reqObj.setReturnFormat(return_format="jsonData")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        term = self.__reqObj.getValue("term")

        # searchTarget = self.__reqObj.getValue("searchTarget")
        # searchOp = self.__reqObj.getValue("searchOp")
        searchName = self.__reqObj.getValue("searchName")
        searchTypeInput = self.__reqObj.getValue("searchType")
        searchType, queryType, inputType, compareType = self.__getSearchType(searchTypeInput)

        logger.info("searchType %r queryType %r compareType %r term %r", searchType, queryType, compareType, term)
        vList = []
        try:
            crs = ChemRefSearch(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            crs.setSearch(queryType, searchType, term, searchName, inputType, compareType)
            _, vList = crs.doSearchAutoComplete()
            logger.debug("\n+ChemRefDataWebAppWorker._chemRefFullSearchAutoCompleteOp() -  term %s for type %s length %d", term, searchType, len(vList))
        except Exception as e:
            logger.exception("Failing with %r", str(e))

        rC.setData(vList)
        return rC

    def _chemRefFullSearchOp(self):
        #
        logger.info("+ChemRefDataWebAppWorker._chemRefFullSearchOp() starting")
        #
        #  Form search target value provided by user -
        searchTarget = self.__reqObj.getValue("searchTarget")
        # static hidden title defined in the input form
        # searchTitle = self.__reqObj.getValue("searchTitle")
        # HTML template path -- usually self.__templatePath -
        appsHtdocsPath = self.__reqObj.getValue("AppHtdocsPath")
        # Form selection option name
        searchName = self.__reqObj.getValue("searchName")
        #
        searchTypeInput = self.__reqObj.getValue("searchType")
        searchType, queryType, inputType, compareType = self.__getSearchType(searchTypeInput)
        #
        logger.info(
            "searchType %r queryType %r searchTarget %r inputType %r compareType %r appsHtdocsPath %r", searchType, queryType, searchTarget, inputType, compareType, appsHtdocsPath
        )
        # enable comma-separated lists or hyphen-separated ranges in search box
        if searchType.startswith("CCD_CC_ID") or searchType.startswith("BIRD_PRD_ID"):
            searchTarget = searchTarget.replace(",", " ")
        elif searchType == "CCD_FORMULA_WEIGHT" or searchType == "CCDIDX_FORMULA_WEIGHT_RANGE":
            searchTarget = searchTarget.replace("-", " ")
        #
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)

        #
        rD = {}
        try:
            crs = ChemRefSearch(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            crs.setSearch(queryType, searchType, searchTarget, searchName, inputType, compareType)
            rD = crs.doSearch()
        except Exception as e:
            logger.exception("Failing with %r", str(e))
        #
        renderBsTable = True
        ##
        #
        if renderBsTable:
            if len(rD) > 0:
                logger.debug("Rendering bootstrap table style with object length %r", len(rD))
                includePath = os.path.join(appsHtdocsPath, "includes")
                crsdp = ChemRefSearchDepictBootstrap(includePath=includePath, verbose=self.__verbose, log=self.__lfh)
                tableDataD = crsdp.doBsTableRenderCollapsable(rD, searchName=searchName)
                if len(tableDataD) > 0:
                    # Handle the JSON conversion under the hood -
                    # rC.set('resultSetTableData', json.dumps(tableDataD))
                    rC.set("resultSetTableData", tableDataD, asJson=True)
                    rC.set("resultSetId", str(1))
                    rC.setStatus(statusMsg="Search completed")
                else:
                    rC.setStatus(statusMsg="No search results")
            else:
                rC.setStatus(statusMsg="No search results")
        #
        else:
            if len(rD) > 0:
                logger.debug("Rendering with object length %r", len(rD))
                includePath = os.path.join(appsHtdocsPath, "includes")
                crsdp = ChemRefSearchDepictBootstrap(includePath=includePath, verbose=self.__verbose, log=self.__lfh)
                oL = crsdp.doAltRenderCollapsable(rD, searchName=searchName)
                if len(oL) > 0:
                    rC.setHtmlList(oL)
                    rC.setStatus(statusMsg="Search completed")
                else:
                    rC.setStatus(statusMsg="No search results")
            else:
                rC.setStatus(statusMsg="No search results")

        return rC

    def __makeIdListReportResponse(self, idCodeList):
        """Prepare response for a report request for the input Id code list."""
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        du = DownloadUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        aTagList = []
        htmlList = []
        webPathList = []
        myIdList = []

        for idCode in idCodeList:
            du.fetchId(idCode)
            repType = du.getIdType(idCode)
            downloadPath = du.getDownloadPath()
            aTagList.append(du.getAnchorTag())
            oL, _, webXyzPath = self.__makeTabularReport(filePath=downloadPath, repositoryType=repType, idCode=idCode, layout="tabs")
            htmlList.extend(oL)
            webPathList.append(webXyzPath)
            myIdList.append(idCode)
        if len(aTagList) > 0:
            rC.setHtmlLinkText('<span class="url-list">View: %s</span>' % ",&nbsp;&nbsp;".join(aTagList))
            rC.setHtmlList(htmlList)
            rC.setStatus(statusMsg="Reports completed")
            rC.set("webPathList", webPathList)
            rC.set("idCodeList", myIdList)
        else:
            rC.setError(errMsg="No corresponding reference file(s)")
            # do nothing

        return rC

    def __makeTabularReport(self, filePath, repositoryType, idCode, layout="tabs"):
        """Internal method to create a tabular report corresponding to the input
        chemical reference definition file.

        layout = tabs|accordion|multiaccordion

        Return data as a list of HTML markup for the section containing the tabular report.
        """
        #
        oL = []
        fileFormat = "cif"
        webXyzPath = None
        #
        logger.info("+ChemRefDataWebAppWorker.__makeTabularReport() target file %s id code %s repository %s", filePath, idCode, repositoryType)

        if filePath is not None and repositoryType is not None and idCode is not None:
            if repositoryType in ["CC", "PRDCC"]:
                rd = ChemCompReport(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                rd.setFilePath(filePath, ccFileFormat=fileFormat, ccId=idCode)
                dd = rd.doReport()
                #
                rdd = ChemRefReportDepictBootstrap(styleObject=ChemCompCategoryStyle(), verbose=self.__verbose, log=self.__lfh)
                oL = rdd.render(dd, style=layout)
                webXyzPath = dd["xyzRelativePath"]
            elif repositoryType in ["PRD", "PRD_FAMILY"]:
                rd = BirdReport(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
                rd.setFilePath(filePath, prdFileFormat=fileFormat, prdId=idCode)
                dd = rd.doReport()
                #
                rdd = ChemRefReportDepictBootstrap(styleObject=PrdCategoryStyle(), verbose=self.__verbose, log=self.__lfh)
                oL = rdd.render(dd, style=layout)
                webXyzPath = dd["xyzRelativePath"]
            if self.__debug:
                logger.info("+ChemRefDataWebAppWorker.__makeTabularReport - generated HTML \n%s", "\n".join(oL))

        return oL, idCode, webXyzPath

    # --------------------------------------------------------------------------------------
    #
    #       Current Admin methods ---
    #
    def _chemRefAdminOps(self):
        """Chemical reference data repository/sandbox level administrative operations."""
        operation = self.__reqObj.getValue("operation")
        logger.info("+ChemRefDataWebAppWorker._chemRefAdminOps() starting with op %s", operation)

        if operation == "updatedatabasebird":
            return self.__chemRefDatabaseUpdateOp(referenceDatabase="BIRD")
        elif operation == "synccvsbird":
            return self.__chemRefSyncCvsOp(repositoryType="BIRD")
        elif operation == "updatedatabasecc":
            return self.__chemRefDatabaseUpdateOp(referenceDatabase="CC")
        elif operation == "synccvscc":
            return self.__chemRefSyncCvsOp(repositoryType="CC")
        elif operation == "updatesupportfiles":
            return self.__chemRefSupportFileUpdateOp()
        elif operation == "updateindexfiles":
            return self.__chemRefIndexFileUpdateOp()
        else:
            return self.__chemRefSyncCvsOp(repositoryType=None)

    def __chemRefSupportFileUpdateOp(self):
        self.__getSession()
        logger.info("+ChemRefDataWebAppWorker._chemRefSupportFileUpdateOp() starting with session %s", self.__sessionPath)

        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)

        ok = self.__chemRefSupportFiles()

        if ok:
            rC.setStatus(statusMsg="Chemical component support file update completed")
        else:
            rC.setError(errMsg="Chemical component support file update failed")

        return rC

    def __chemRefSupportFiles(self):
        """Update chemical component definition file idList, pathList, concatenated dictionary,
        serialized dictionary, dictionary search index, and several Python serialized object/index files.
        """
        logger.info("Starting")
        try:
            mu = ChemRefDataMiscUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok1 = mu.updateChemCompSupportFiles()
            ok2 = mu.updateChemCompPySupportFiles()
            ok3 = mu.updatePrdSupportFiles()
            ok = ok1 and ok2 and ok3
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failure in __chemRefSupportFiles")
            ok = False
        return ok

    def __chemRefIndexFileUpdateOp(self):
        self.__getSession()
        logger.info("+ChemRefDataWebAppWorker._chemRefIndexFileUpdateOp() starting with session %s", self.__sessionPath)

        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)

        ok = self.__chemRefIndexFiles()

        if ok:
            rC.setStatus(statusMsg="Chemical component index file update completed")
        else:
            rC.setError(errMsg="Chemical component index file update failed")

        return rC

    def __chemRefIndexFiles(self):
        """Update Python serialized object and index files."""
        logger.info("Starting")
        ok = False
        try:
            mu = ChemRefDataMiscUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
            ok = mu.updateChemCompPySupportFiles()
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failure in __chemRefIndexFiles")

        return ok

    def __chemRefSyncCvsOp(self, repositoryType=None):
        """Synchronize the reference data sandbox areas with the reference CVS repository."""
        logger.info("+ChemRefDataWebAppWorker._chemRefSyncCvsOp() starting")

        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        crdu = ChemRefDataCvsUtils(self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        if repositoryType in ["CC"]:
            ok, textList = crdu.syncChemComp()
        elif repositoryType in ["PRD", "PRD_FAMILY", "PRDCC", "BIRD"]:
            ok, textList = crdu.syncBird()
        else:
            ok = False
            textList = ["Unknown reference respository"]

        logger.info("+ChemRefDataWebAppWorker._chemRefSyncCvsOp() repository %s completion status is %r", repositoryType, ok)
        #
        if ok:
            rC.setStatus(statusMsg="Repository SYNC completed")
        else:
            if self.__verbose:
                logger.info("+ChemRefDataWebAppWorker._chemRefSyncCvsOp() diagnostics %r", textList)
            hL = []
            for tt in textList:
                tL = tt.split("\n")
                hL.append("<br />".join(tL))

            rC.setError(errMsg="Repository SYNC completed with diagnostics  <br />" + "<br />".join(hL))

        return rC

    def __chemRefDatabaseUpdateOp(self, referenceDatabase="CC"):
        self.__getSession()
        logger.info("+ChemRefDataWebAppWorker._chemRefDatabaseUpdateOp() starting with session %s", self.__sessionPath)

        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)

        crdbu = ChemRefDataDbUtils(self.__reqObj, verbose=self.__debug, log=self.__lfh)
        if referenceDatabase in ["CC"]:
            ok = crdbu.loadChemCompMulti()
        elif referenceDatabase in ["BIRD", "PRD", "PRD_FAMILY"]:
            ok = crdbu.loadBird()
        else:
            logger.error("referenceDatabase %s unknown", referenceDatabase)
            ok = False

        if ok:
            rC.setStatus(statusMsg="%s database load completed" % referenceDatabase)
        else:
            rC.setError(errMsg="Database load failed")

        return rC

    #
    #  supporting internal methods --
    #
    def X_getFileText(self, filePath):
        self.__reqObj.setReturnFormat(return_format="text")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        rC.setTextFile(filePath)
        return rC

    def _newSessionOp(self):
        logger.info("+AnnTasksWebAppWorker._newSessionOp() starting")

        self.__getSession(forceNew=True)
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        sId = self.__reqObj.getSessionId()
        if len(sId):
            rC.setHtmlText("Session id %s created." % sId)
        else:
            rC.setError(errMsg="No session created")

        return rC

    def __getSession(self, forceNew=False):
        """Join existing session or create new session as required."""
        #
        sessionId = self.__reqObj.getSessionId()
        logger.info("__getSession() - Starting with session %s", sessionId)
        #
        self.__sObj = self.__reqObj.newSessionObj(forceNew=forceNew)

        self.__sessionId = self.__sObj.getId()
        self.__sessionPath = self.__sObj.getPath()
        self.__rltvSessionPath = self.__sObj.getRelativePath()

        logger.info("__getSession() - Leaving with session %s", self.__sessionId)
        logger.info("__getSession() - Leaving session path %s", self.__sessionPath)
        logger.info("------------------------------------------------------")

    def __isFileUpload(self, fileTag="file"):
        """Generic check for the existence of request paramenter "file"."""
        # Gracefully exit if no file is provide in the request object -
        fs = self.__reqObj.getRawValue(fileTag)
        if sys.version_info[0] < 3:
            if (fs is None) or (isinstance(fs, types.StringType)):  # pylint: disable=no-member
                return False
        else:
            if (fs is None) or isinstance(fs, (str, bytes)):
                return False

        return True

    def __uploadFile(self, fileTag="file", fileTypeTag="filetype"):
        #
        #
        logger.info("+ChemRefDataWebApp.__uploadFile() - file upload starting")

        #
        # Copy upload file to session directory -
        try:
            fs = self.__reqObj.getRawValue(fileTag)
            # fNameInput = str(fs.filename).lower()
            fNameInput = str(fs.filename)
            #
            # Need to deal with some platform issues -
            #
            if fNameInput.find("\\") != -1:
                # likely windows path -
                fName = ntpath.basename(fNameInput)
            else:
                fName = os.path.basename(fNameInput)

            #

            logger.info("+ChemRefDataWebApp.__loadDataFileStart() - upload file %s", fs.filename)
            logger.info("+ChemRefDataWebApp.__loadDataFileStart() - base file   %s", fName)
            #
            # Store upload file in session directory -

            fPathAbs = os.path.join(self.__sessionPath, fName)
            ofh = open(fPathAbs, "wb")
            ofh.write(fs.file.read())
            ofh.close()
            self.__reqObj.setValue("UploadFileName", fName)
            self.__reqObj.setValue("filePath", fPathAbs)
            logger.info("+ChemRefDataWebApp.__uploadFile() Uploaded file %s", str(fName))
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("+ChemRefDataWebApp.__uploadFile() File upload processing failed for %s", str(fs.filename))

            return False
        #
        # Verify file name
        #
        fs = self.__reqObj.getRawValue(fileTag)
        fmt = self.__reqObj.getValue(fileTypeTag)
        fType = fmt.lower()
        # fNameInput=str(fs.filename).strip().lower()
        fNameInput = str(fs.filename).strip()
        #
        # Need to deal with some platform issues -
        #
        if fNameInput.find("\\") != -1:
            # likely windows path -
            fName = ntpath.basename(fNameInput)
        else:
            fName = os.path.basename(fNameInput)
        #
        #
        if fName.startswith("rcsb"):
            fId = fName[:10]
        elif fName.startswith("d_"):
            fId = fName[:8]
        else:
            fId = "000000"
            logger.info("+ChemRefDataWebApp.__uploadFile() using default identifier for %s", str(fName))

        self.__reqObj.setValue("identifier", fId)
        self.__reqObj.setValue("fileName", fName)
        #
        if fType in ["cif", "cifeps", "pdb"]:
            self.__reqObj.setValue("fileType", fType)
        else:
            self.__reqObj.setValue("fileType", "unknown")

        logger.info("+ChemRefDataWebApp.__uploadFile() identifier %s", self.__reqObj.getValue("identifier"))
        logger.info("+ChemRefDataWebApp.__uploadFile() UploadFileType  %s", self.__reqObj.getValue("UploadFileType"))
        logger.info("+ChemRefDataWebApp.__uploadFile() UploadFileName  %s", self.__reqObj.getValue("UploadFileName"))
        return True

    def _chemRefEditorOps(self):
        """Launch chemical component editor based on id code."""
        self.__getSession()
        self.__reqObj.setReturnFormat(return_format="json")
        rC = ResponseContent(reqObj=self.__reqObj, verbose=self.__verbose, log=self.__lfh)
        #
        idCode = self.__reqObj.getValue("idcode")
        logger.info("+ChemRefDataWebAppWorker._chemRefEditorOps() idCode %r", idCode)
        #
        if not idCode:
            rC.setError(errMsg="Please input CC ID")
            return rC
        #
        if len(idCode) > 5:
            rC.setError(errMsg="Invalid CC ID")
            return rC
        #
        filePath = self.__crPI.getFilePath(idCode=idCode)
        if not filePath:
            rC.setError(errMsg="Invalid CC ID")
        else:
            localFilePath = os.path.join(self.__sessionPath, idCode.upper() + ".cif")
            shutil.copyfile(filePath, localFilePath)
            rC.setStatus(statusMsg="Load completed")
            #
            relStatus = ""
            try:
                cifObj = mmCIFUtil(filePath=localFilePath)
                relStatus = cifObj.GetSingleValue("chem_comp", "pdbx_release_status").strip().upper()
            except:  # noqa: E722 pylint: disable=bare-except
                logger.exception("Failure in in _ChemRefEditorOps")
            #
            myD = {}
            myD["sessionid"] = self.__sessionId
            myD["instanceid"] = idCode.upper()
            myD["processing_site"] = self.__cI.get("SITE_NAME").upper()
            myD["urlcifpath"] = os.path.join(self.__rltvSessionPath, idCode.upper() + ".cif")
            if relStatus == "REL":
                myD["ccmodel"] = "compCompModelRel"
            else:
                myD["ccmodel"] = "compCompModel"
            #
            htmlText = self.__processTemplate("templates/cc_edit_tmplt.html", myD)
            #
            htmlFilePath = os.path.join(self.__sessionPath, idCode.upper() + ".html")
            ofh = open(htmlFilePath, "w")
            ofh.write(htmlText + "\n")
            ofh.close()
            #
            rC.setLocation(os.path.join(self.__rltvSessionPath, idCode.upper() + ".html"))
        #
        return rC

    def __processTemplate(self, fn, parameterDict=None):
        """Read the input HTML template data file and perform the key/value substitutions in the
        input parameter dictionary.

        :Params:
            ``parameterDict``: dictionary where
            key = name of subsitution placeholder in the template and
            value = data to be used to substitute information for the placeholder

        :Returns:
            string representing entirety of content with subsitution placeholders now replaced with data
        """
        if parameterDict is None:
            parameterDict = {}

        tPath = self.__reqObj.getValue("TemplatePath")
        fPath = os.path.join(tPath, fn)
        with open(fPath, "r") as ifh:
            sIn = ifh.read()
        return sIn % parameterDict


def main():
    sTool = ChemRefDataWebApp()
    d = sTool.doOp()
    for k, v in d.items():
        sys.stdout.write("Key - %s  value - %r\n" % (k, v))


if __name__ == "__main__":
    main()



##
# File:  OSVersion.py
# Date:  19-Nove-2021  E. Peisach
#
# Updated:
#
##
"""
Utility to identify operating system particulars
"""
__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "peisach@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import distro


class OSVersion(object):
    __osdata = None

    def __getOsInfo(self):
        if self.__osdata is None:
            self.__osdata = distro.info()

    def IsRhelLike(self):
        """Determines if O/S is like redhat.  Works with centos, fedora, and rocky linux"""

        self.__getOsInfo()

        like = self.__osdata.get("like", "")

        if "rhel" in like:
            return True

        return False

    def IsRhel8Like(self):
        """Returns True if systm like RHEL 8"""

        if self.IsRhelLike():
            major = self.__osdata.get("version_parts", {}).get("major")
            if major == "8":
                return True

        return False


def main():
    osv = OSVersion()
    print(osv.IsRhel8Like())


if __name__ == "__main__":
    main()



##
# File:  ReportUtils.py
# Date:  14-Jun-2017 jdw
#
# Updates:
#  14-Jun-2017  jdw refactored common modules
##
"""
Common methods shared by report modules -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"


import sys
import logging

from wwpdb.utils.oe_util.oedepict.OeDepict import OeDepict
from wwpdb.utils.oe_util.build.OeChemCompIoUtils import OeChemCompIoUtils

logger = logging.getLogger(__name__)


class ReportUtils(object):
    """
    Common methods shared by report modules -

    """

    def __init__(self, verbose=False, log=sys.stderr):
        """PRD report generator.

        :param `verbose`:  boolean flag to activate verbose logging.
        :param `log`:      stream for logging.

        """
        self.__verbose = verbose
        self.__lfh = log
        #

    def coordinatesExist(self, cD):
        """Return a tuple of booleans (exptl xyz exist, ideal xyz exist)"""
        hasExpt = False
        hasIdeal = False
        try:
            if "chem_comp" in cD:
                d = cD["chem_comp"][0]
                logger.info("d is %r", d.items())
                if "_chem_comp.pdbx_model_coordinates_missing_flag" in d:
                    hasExpt = d["_chem_comp.pdbx_model_coordinates_missing_flag"] == "N"
                if "_chem_comp.pdbx_ideal_coordinates_missing_flag" in d:
                    hasIdeal = d["_chem_comp.pdbx_ideal_coordinates_missing_flag"] == "N"
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed with cD %r", cD.items())

        logger.info("exptl %r ideal %r", hasExpt, hasIdeal)
        return hasExpt, hasIdeal

    def makeImagesFromPathList(self, pathList=None, imagePath="image.png", rows=1, cols=1, imageX=500, imageY=500):
        """Fetch, read, build, and depict the OE molecule from input path list of chemical definitions."""
        logger.info("Starting")
        if pathList is None:
            pathList = []
        try:
            oeMolTitleList = self.__makeOeMols(pathList=pathList)
            oed = OeDepict(verbose=self.__verbose, log=self.__lfh)
            oed.setMolTitleList(oeMolTitleList)
            oed.setDisplayOptions(imageX=imageX, imageY=imageY, labelAtomName=True, labelAtomCIPStereo=True, labelAtomIndex=False, labelBondIndex=False, bondDisplayWidth=0.5)
            oed.setGridOptions(rows=rows, cols=cols)
            oed.prepare()
            oed.write(imagePath)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failing for pathList %r", pathList)
        return False

    def __makeOeMols(self, pathList=None):
        """Create OE molecules from the input chem comp definition path list."""
        if pathList is None:
            pathList = []
        try:
            oemList = []
            idList = []
            oeU = OeChemCompIoUtils(verbose=self.__verbose, log=self.__lfh)
            oemList = oeU.getFromPathList(pathList, use3D=False, coordType="model")
            for oem in oemList:
                title = oem.getTitle()
                idList.append(title)
                logger.info("Title              = %s", title)
                logger.info("SMILES (canonical) = %s", oem.getCanSMILES())
                logger.info("SMILES (isomeric)  = %s", oem.getIsoSMILES())
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failing for pathList %r", pathList)

        return list(zip(idList, oemList, idList))
